import os 
import tqdm
import shutil
from .. import haven_utils as hu 
from .. import haven_jupyter as hj


def to_dropbox(exp_list, savedir_base, dropbox_path, access_token, zipname):
    """[summary]
    
    Parameters
    ----------
    exp_list : [type]
        [description]
    savedir_base : [type]
        [description]
    dropbox_path : [type]
        [description]
    access_token : [type]
        [description]
    """
    # zip files 
    exp_id_list = [hu.hash_dict(exp_dict) for exp_dict in exp_list]
    src_fname = os.path.join(savedir_base, zipname)
    out_fname = os.path.join(dropbox_path, zipname)
    zipdir(exp_id_list, savedir_base, src_fname)

    upload_file_to_dropbox(src_fname, out_fname, access_token)
    print('saved: https://www.dropbox.com/home/%s' % out_fname)

    
def upload_file_to_dropbox(src_fname, out_fname, access_token):
    import dropbox
    dbx = dropbox.Dropbox(access_token)
    try:
        dbx.files_delete_v2(out_fname)
    except:
        pass
    # with open(src_fname, 'rb') as f:
    #     dbx.files_upload(f.read(), out_fname)

    upload(access_token=access_token,
           file_path=src_fname,
           target_path=out_fname)


def upload(
    access_token,
    file_path,
    target_path,
    timeout=900,
    chunk_size=4 * 1024 * 1024,
):
    import os
    import dropbox
    import tqdm
    dbx = dropbox.Dropbox(access_token, timeout=timeout)
    with open(file_path, "rb") as f:
        file_size = os.path.getsize(file_path)
        chunk_size = 4 * 1024 * 1024
        if file_size <= chunk_size:
            print(dbx.files_upload(f.read(), target_path))
        else:
            with tqdm.tqdm(total=file_size, desc="Uploaded") as pbar:
                upload_session_start_result = dbx.files_upload_session_start(
                    f.read(chunk_size)
                )
                pbar.update(chunk_size)
                cursor = dropbox.files.UploadSessionCursor(
                    session_id=upload_session_start_result.session_id,
                    offset=f.tell(),
                )
                commit = dropbox.files.CommitInfo(path=target_path)
                while f.tell() < file_size:
                    if (file_size - f.tell()) <= chunk_size:
                        print(
                            dbx.files_upload_session_finish(
                                f.read(chunk_size), cursor, commit
                            )
                        )
                    else:
                        dbx.files_upload_session_append(
                            f.read(chunk_size),
                            cursor.session_id,
                            cursor.offset,
                        )
                        cursor.offset = f.tell()
                    pbar.update(chunk_size)
    print('uploaded!')


def zipdir(exp_id_list, savedir_base, src_fname, add_jupyter=True, verbose=1, 
           fname_list=None, dropbox_path='/shared', access_token=None):
    import zipfile
    zipf = zipfile.ZipFile(src_fname, 'w', zipfile.ZIP_DEFLATED)

    # ziph is zipfile handle
    if add_jupyter:
        abs_path = os.path.join(savedir_base, 'results.ipynb')
        hj.create_jupyter(fname=abs_path, 
                        savedir_base='results/', overwrite=False, print_url=False,
                        create_notebook=True)
        
        rel_path = 'results.ipynb'
        zipf.write(abs_path, rel_path)
        os.remove(abs_path)

    n_zipped = 0
    if verbose:
        tqdm_bar = tqdm.tqdm
    else:
        tqdm_bar = lambda x: x
    
    fname_all = ['score_list.pkl', "exp_dict.json"] 
    if isinstance(fname_list, list):
        fname_all += fname_list

    for exp_id in tqdm_bar(exp_id_list):
        if not os.path.isdir(os.path.join(savedir_base, exp_id)):
            continue
            
        for fname in fname_all:
            abs_path = os.path.join(savedir_base, exp_id, fname)
            rel_path = os.path.join( "results", exp_id, fname)
            if os.path.exists(abs_path):
                zipf.write(abs_path, rel_path)

        n_zipped += 1

    zipf.close()
    if verbose:
        print('Zipped: %d/%d exps in %s' % (n_zipped, len(exp_id_list), src_fname))

    if access_token is not None and access_token != '':
        out_fname = os.path.join(dropbox_path, src_fname)
        upload_file_to_dropbox(src_fname, out_fname, access_token)
        print('saved: https://www.dropbox.com/home/%s' % out_fname)


import copy
import glob
import os
import sys
import pprint
from itertools import groupby
from textwrap import wrap
import numpy as np
import pandas as pd
import pylab as plt
import tqdm

from .. import haven_jobs as hjb
from .. import haven_utils as hu
from .. import haven_share as hd

def get_score_df(exp_list, savedir_base, filterby_list=None, columns=None,
                 score_columns=None,
                 verbose=True, wrap_size=8, hparam_diff=0, flatten_columns=True,
                 show_meta=True, show_max_min=True, add_prefix=False,
                 score_list_name='score_list.pkl', in_latex_format=False, avg_across=None):
    """Get a table showing the scores for the given list of experiments 

    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    columns : list, optional
        a list of columns you would like to display, by default None
    savedir_base : str, optional
        A directory where experiments are saved
        
    Returns
    -------
    DataFrame
        a dataframe showing the scores obtained by the experiments

    Example
    -------
    >>> from haven import haven_results as hr
    >>> savedir_base='../results/isps/'
    >>> exp_list = hr.get_exp_list(savedir_base=savedir_base, 
    >>>                            filterby_list=[{'sampler':{'train':'basic'}}])
    >>> df = hr.get_score_df(exp_list, savedir_base=savedir_base, columns=['train_loss', 'exp_id'])
    >>> print(df)
    """
    if len(exp_list) == 0:
        if verbose:
            print('exp_list is empty...')
        return
    exp_list = hu.filter_exp_list(exp_list, filterby_list, savedir_base=savedir_base, verbose=verbose)

    # aggregate results
    result_list = []
    for exp_dict in exp_list:
        result_dict = {}
        exp_id = hu.hash_dict(exp_dict)
        if avg_across is not None:
            tmp_dict = copy.deepcopy(exp_dict)
            del tmp_dict[avg_across]
            result_dict['_' + avg_across] = hu.hash_dict(tmp_dict)

        savedir = os.path.join(savedir_base, exp_id)
        score_list_fname = os.path.join(savedir, score_list_name)
        exp_dict_fname = os.path.join(savedir, "exp_dict.json")

        hparam_columns = columns or list(exp_dict.keys())
        for k in hparam_columns:
            if add_prefix:
                k_new = "(hparam) " + k
            else:
                k_new = k
            result_dict[k_new] = exp_dict[k]

        if os.path.exists(score_list_fname) and show_meta:
            result_dict['started_at'] = hu.time_to_montreal(exp_dict_fname)
            result_dict['creation_time'] = os.path.getctime(exp_dict_fname)
        else:
            result_dict['creation_time'] = -1

        if show_meta:
            result_dict["exp_id"] = exp_id

        if flatten_columns:
            result_dict = hu.flatten_column(result_dict)

        hparam_columns = list(result_dict.keys())

        if not os.path.exists(score_list_fname):
            if verbose:
                print('%s: %s is missing' % (exp_id, score_list_name))
            
        else:
            score_list = hu.load_pkl(score_list_fname)
            score_df = pd.DataFrame(score_list)
            metric_columns = score_columns or score_df.columns
            if len(score_list):
                for k in metric_columns:
                    v = np.array(score_df[k])
                    if 'float' in str(v.dtype):
                        v = v[~np.isnan(v)]

                    if len(v):
                        if add_prefix:
                            k_new = "(metric) " +k
                        else:
                            k_new = k

                        if "float" in str(v.dtype):
                            result_dict[k_new] = v[-1]
                            if show_max_min:
                                result_dict[k_new+' (max)'] = v.max()
                                result_dict[k_new+' (min)'] = v.min()
                                
                        else:
                            result_dict[k_new] = v[-1]
        

        result_list += [result_dict]

    # create table
    df = pd.DataFrame(result_list)
    metric_columns = [c for c in result_dict if c not in hparam_columns]
    # print(avg_across)
    if avg_across is not None:
        df_avg = df.groupby('_' + avg_across)
        df_tmp = df_avg[hparam_columns].first().join(df_avg[metric_columns].agg([np.mean, np.std]))
        del df_tmp['_' + avg_across]
        df_tmp = df_tmp.reset_index()
        del df_tmp['_' + avg_across]
        # del df_tmp[avg_across]
        df = df_tmp 

    df = df.sort_values(by='creation_time' )
    del df['creation_time']

    # wrap text for prettiness
    df = hu.pretty_print_df(df)

    if hparam_diff > 0 and len(df) > 1:
        cols = hu.get_diff_columns(df, min_threshold=hparam_diff, max_threshold='auto')
        df = df[cols]

    df =  hu.sort_df_columns(df)
    if in_latex_format:
        return df.to_latex(index=False)
    return df

def get_score_lists(exp_list, savedir_base, filterby_list=None, verbose=True,
                    score_list_name='score_list.pkl'):
    """[summary]
    
    Parameters
    ----------
    exp_list : [type]
        A list of experiments, each defines a single set of hyper-parameters
    savedir_base : [type]
        [description]
    filterby_list : [type], optional
        [description], by default None
    
    Returns
    -------
    [type]
        [description]
    Example
    -------
    >>> from haven import haven_results as hr
    >>> savedir_base='../results/isps/'
    >>> exp_list = hr.get_exp_list(savedir_base=savedir_base, 
    >>>                            filterby_list=[{'sampler':{'train':'basic'}}])
    >>> lists_of_score_lists = hr.get_score_lists(exp_list, savedir_base=savedir_base, columns=['train_loss', 'exp_id'])
    >>> print(lists_of_score_lists)
    """
    if len(exp_list) == 0:
        if verbose:
            print('exp_list is empty...')
        return

    exp_list = hu.filter_exp_list(exp_list, filterby_list, savedir_base=savedir_base, verbose=verbose)
    score_lists = []

    # aggregate results
    for exp_dict in exp_list:
        exp_id = hu.hash_dict(exp_dict)
        savedir = os.path.join(savedir_base, exp_id)

        score_list_fname = os.path.join(savedir, score_list_name)
        if not os.path.exists(score_list_fname):
            if verbose:
                print('%s: missing %s' % (exp_id, score_list_name))
            continue
        
        else:
            score_lists += [hu.load_pkl(score_list_fname)]
    
    return score_lists


def get_exp_list_df(exp_list, filterby_list=None, columns=None, verbose=True):
    """Get a table showing the configurations for the given list of experiments 

    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    columns : list, optional
        a list of columns you would like to display, by default None
        
    Returns
    -------
    DataFrame
        a dataframe showing the scores obtained by the experiments

    Example
    -------
    >>> from haven import haven_results as hr
    >>> savedir_base='../results/isps/'
    >>> exp_list = hr.get_exp_list(savedir_base=savedir_base, 
    >>>                            filterby_list=[{'sampler':{'train':'basic'}}])
    >>> df = hr.get_exp_list_df(exp_list, columns=['train_loss', 'exp_id'])
    >>> print(df)
    """
    if len(exp_list) == 0:
        if verbose:
            print('exp_list is empty...')
        return

    exp_list = hu.filter_exp_list(exp_list, filterby_list, verbose=verbose)

    # aggregate results
    result_list = []
    for exp_dict in exp_list:
        result_dict = {}

        exp_id = hu.hash_dict(exp_dict)
        result_dict["exp_id"] = exp_id

        for k in exp_dict:
            result_dict[k] = exp_dict[k]

        result_list += [result_dict]

    df = pd.DataFrame(result_list)

    if columns:
        df = df[[c for c in columns if c in df.columns]]

    return df


import pandas as pd
import pylab as plt 


def merge_pdfs(fname_list, output_name="output.pdf"):
    from PyPDF2 import PdfFileReader, PdfFileMerger
    import PyPDF2
    import img2pdf

    from pdfrw import PdfReader, PdfWriter

    writer = PdfWriter()
    for inpfn in fname_list:
        decrypt_pdf(inpfn, output_name)
        writer.addpages(PdfReader(output_name).pages)
    writer.write(output_name)
    

def decrypt_pdf(input_name, output_name):
    import pikepdf

    pdf = pikepdf.open(input_name)
    pdf.save(output_name)

def images_to_pdf(fname_list, output_name="output.pdf"):
    from fpdf import FPDF
    pdf = FPDF()
    # imagelist is the list with all image filenames
    for image in fname_list:
        pdf.add_page()
        pdf.image(image,x=0,y=0,w=610,h=297)
    pdf.output(output_name, "F")



import os, subprocess, time
import argparse, pprint

 
JOB_CONFIG = """
image: registry.console.elementai.com/{ACCOUNT}/ssh
data:
    - {ACCOUNT}.home:/mnt/home
    - {ACCOUNT}.results:/mnt/results
    - {ACCOUNT}.datasets:/mnt/datasets
    - {ACCOUNT}.public:/mnt/public
    - {ACCOUNT}.private:/mnt/private
resources:
    cpu: 8
    mem: 8
    gpu: 1
interactive: true
command:
    - /tk/bin/start.sh
    - /bin/bash
    - -c
    - jupyter notebook --notebook-dir='/mnt' --ip=0.0.0.0 --port=8080 --no-browser --NotebookApp.token='' --NotebookApp.custom_display_url=https://${EAI_JOB_ID}.job.console.elementai.com --NotebookApp.disable_check_xsrf=True --NotebookApp.allow_origin='*'
        """

def get_interactive_job_id(account):
    command = ('eai job ls --account %s --fields state '
                    '--fields id,interactive | '
                    'grep RUNNING | grep true' % account)
    print(command)
    out = os.popen(command).read()
    print('out: %s' % out)

    if out == '':
        return None 

    job_id = out.split()[1]
    return job_id

def kill_job(job_id):
    os.popen('eai job kill %s ' % job_id).read()

def wait_until_state(job_id, required_state):
    elapsed = 0
    max_elapsed = 60

    state = get_job_state(job_id)
    while state != required_state and elapsed < max_elapsed:
        print('%s interactive job: %s (Elapsed %s/%s)' % 
             (state, job_id, elapsed, max_elapsed))
        time.sleep(2)
        state = get_job_state(job_id)
        elapsed += 2

def launch_interactive_job(role):
    command = ('eai job submit --no-header '
               '-f job.yml '
               '--role %s' % (role))

    print(command)
    out = os.popen(command).read()
    print('out: %s' % out)

    return out.split(' ')[0]

def get_job_state(job_id):
    command = ('eai job  info %s '  
                 '| grep state:' % job_id)
    state = os.popen(command).read()
    return state.split(':')[-1].replace('\n','').strip(' ')

def do_port_forwarding(job_id):
    out = os.popen('kill $(lsof -ti:2222)').read()
    command = ('eai job port-forward %s 2222' % (job_id))

    print(command)
    out = os.popen(command).read()
    print('out: %s' % out)

    return out


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    # example user: 'eai.issam'
    parser.add_argument('-a', '--account', required=True)
    # example role: 'eai.issam.admin'
    parser.add_argument('-r', '--role', required=True)
    args = parser.parse_args()

    # define job.yml
    JOB_CONFIG = JOB_CONFIG.replace('{ACCOUNT}', args.account)
    pprint.pprint(JOB_CONFIG)
    print('role:', args.role)
    with open('job.yml', 'w') as outfile:
        outfile.write(JOB_CONFIG)

    # make sure no interactive job is running
    job_id = get_interactive_job_id(args.account)
    if job_id == None or job_id == '':
        out = 'y'
    else:
        # interactive job exists
        print('RUNNING interactive job:', job_id)
        out = input('Restart? (y/n) ')
        if out == 'y':
            # kill job until dead
            kill_job(job_id)
            wait_until_state(job_id, required_state='CANCELLED')

    if out == 'y':
        # run new interactive job until it runs
        job_id = launch_interactive_job(role=args.role)
        if job_id == None or job_id == '':
            print('No job was launched')
        else:
            wait_until_state(job_id, required_state='RUNNING')

        # launch ssh
        # do_port_forwarding(job_id)

    # help commands
    print()
    print('Helpers:\n========\n')
    print('eai job port-forward %s 2222' % job_id)
    print('workstation: ssh -p 2222 toolkit@localhost -v')
    print('jupyter url: https://%s.job.console.elementai.com/' % job_id)

    if os.path.exists('job.yml'):
        os.remove('job.yml')




from .. import haven_utils
from .. import haven_results as hr
from .. import haven_utils as hu
from .. import haven_share as hd

import os
import pprint
import json
import copy
import pprint
import pandas as pd

try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except:
    print('widgets not available...')


def plots_tab(self, output):
    llegend_list = widgets.SelectMultiple(options=self.rm.exp_params,
                        value=self.vars.get('legend_list', [self.rm.exp_params[0]]))
    # widgets.Text(
    #     value=str(self.vars.get('legend_list', '[model]')),
    #     description='legend_list:',
    #     disabled=False
    # )
    llegend_format = widgets.Text(
        value=str(self.vars.get('legend_format', '')),
        description='',
        disabled=False
    )
    ltitle_format = widgets.Text(
        value=str(self.vars.get('title_format', '')),
        description='',
        disabled=False
    )

    lcmap = widgets.Text(
        value=str(self.vars.get('cmap', 'jet')),
        description='cmap:',
        layout=self.layout_dropdown,
        disabled=False
    )

    llog_metric_list = widgets.Text(
        value=str(self.vars.get('log_metric_list', '[train_loss]')),
        description='log_metric_list:',
        disabled=False
    )

    metrics_list = [k for k in self.rm_original.score_keys if k is not 'None']

    metrics_txt = widgets.Label(value="Metrics:", 
                                layout=self.layout_label,)
    t_y_metric =  widgets.SelectMultiple(options=metrics_list,
                         value=self.vars.get('y_metrics', [metrics_list[0]]))
    # t_y_metric 

    # t_y_metric = widgets.Text(
    #     value=str(self.vars.get('y_metrics', 'train_loss')),
    #     description='y_metrics:',
    #     disabled=False
    # )

    d_x_metric_columns = widgets.Dropdown(
                options=metrics_list,
                value=self.vars.get('x_metric', metrics_list[0]),
                layout=widgets.Layout(width='300px'),
                disabled=False,
            )

    t_groupby_list = widgets.SelectMultiple(options=['None'] + self.rm.exp_params,
                    value=self.vars.get('groupby_list', ['None']),)
    # widgets.Text(
    #     value=str(self.vars.get('groupby_list')),
    #     description='groupby_list:',
    #     disabled=False
    # )

    t_mode = widgets.Dropdown(
                options=['line', 'bar'],
                value='line',
                layout=widgets.Layout(width='300px'),
                disabled=False,
            )

    t_bar_agg = widgets.Dropdown(
                options=['last', 'max', 'mean'],
                value='last',
                layout=widgets.Layout(width='300px'),
                disabled=False,
            )

    t_title_list = widgets.SelectMultiple(options=self.rm.exp_params,
                    value=self.vars.get('title_list', ['dataset']),)
    # widgets.Text(
    #     value=str(self.vars.get('title_list', 'dataset')),
    #     description='title_list:',
    #     disabled=False
    # )

    d_style = widgets.Dropdown(
        options=['False', 'True'],
        value='False',
        description='Interactive:',
        layout=self.layout_dropdown,
        disabled=False,
    )

    d_avg_across_columns = widgets.Dropdown(
                options=['None'] + self.rm.exp_params,
                value='None',
                layout=widgets.Layout(width='300px'),
                disabled=False,
            )

    bdownload = widgets.Button(description="Download Plots",
                               layout=self.layout_button)
    bdownload_out = widgets.Output(layout=self.layout_button)

    def on_download_clicked(b):
        fname = 'plots.pdf'
        from matplotlib.backends.backend_pdf import PdfPages
        import matplotlib.pyplot as plt

        pp = PdfPages(fname)
        for fig in self.rm_original.fig_list:
            fig.savefig(pp, format='pdf')
        pp.close()

        bdownload_out.clear_output()

        with bdownload_out:
            display(FileLink(fname, result_html_prefix="Download: "))

    bdownload.on_click(on_download_clicked)

    h1 = widgets.Label(value="Y-axis Metrics:", 
                                    layout=widgets.Layout(width='340px'),)
    h11 = widgets.Label(value="X-axis Metric", 
                                    layout=widgets.Layout(width='340px'),)
    h2 = widgets.Label(value="Legend:", 
                                    layout=widgets.Layout(width='340px'),)
    h22 = widgets.Label(value="Format:", 
                                    layout=widgets.Layout(width='340px'),)
    h3 = widgets.Label(value="Title:", 
                                    layout=widgets.Layout(width='340px'),)
    h33 = widgets.Label(value="Format:", 
                                    layout=widgets.Layout(width='340px'),)

    h4 = widgets.Label(value="GroupBy:", 
                                    layout=widgets.Layout(width='340px'),)
    h44 = widgets.Label(value="", 
                                    layout=widgets.Layout(width='340px'),)
    h5 = widgets.Label(value="Plot Mode:", 
                                    layout=widgets.Layout(width='320px'),)
    h55 = widgets.Label(value="Plot Agg", 
                                    layout=widgets.Layout(width='320px'),)
    space = widgets.Label(value="", 
                                    layout=widgets.Layout(width='300px'),)
    brefresh = widgets.Button(description="Display Plot")
    d_avg_across_txt = widgets.Label(value="avg_across:",)
    button = widgets.VBox([

                          widgets.HBox([h1,  h2, h3, h4,  ]),
                           widgets.HBox([t_y_metric,  llegend_list, t_title_list, t_groupby_list,]),

                           widgets.HBox([h11,  h5, h55, d_avg_across_txt, ]),
                           widgets.HBox([d_x_metric_columns,  t_mode, t_bar_agg, d_avg_across_columns,  ]),
                        #    widgets.HBox([ d_avg_across_txt, d_avg_across_columns,  ]),
                           widgets.HBox([brefresh, bdownload, bdownload_out,]),])

    # button = widgets.VBox([widgets.HBox([brefresh, bdownload, bdownload_out]),
    #                        widgets.HBox([t_y_metric,  d_x_metric_columns]),
    #                        widgets.HBox([t_title_list, d_style]),
                          
    #                        widgets.HBox([t_groupby_list, llegend_list, ]),
    #                        widgets.HBox([t_mode, t_bar_agg]),
    #                        widgets.HBox([ltitle_format, llegend_format]),
    #                        widgets.HBox([d_avg_across_columns]),

    #                        ])

    output_plot = widgets.Output()

    def on_clicked(b):
        if d_style.value == 'True':
            from IPython import get_ipython
            ipython = get_ipython()
            ipython.magic("matplotlib widget")
        output_plot.clear_output()
        with output_plot:
            self.update_rm()

            self.vars['y_metrics'] = list(t_y_metric.value)
            self.vars['x_metric'] = d_x_metric_columns.value

            w, h = 10, 5
            if len(self.vars['y_metrics']) > 1:
                figsize = (2*int(w), int(h))
                self.vars['figsize'] = figsize
            else:
                self.vars['figsize'] = (int(w), int(h))

            self.vars['legend_list'] = list(llegend_list.value)
            self.vars['legend_format'] = llegend_format.value
            self.vars['log_metric_list'] = hu.get_list_from_str(
                llog_metric_list.value)
            self.vars['groupby_list'] = list(
                t_groupby_list.value)
            self.vars['mode'] = t_mode.value
            self.vars['title_list'] = list(t_title_list.value)
            self.vars['bar_agg'] = t_bar_agg.value
            self.vars['title_format'] = ltitle_format.value
            self.vars['cmap'] = lcmap.value
            self.vars['avg_across'] = d_avg_across_columns.value

            avg_across_value = self.vars['avg_across']
            if avg_across_value == "None":
                avg_across_value = None

            self.rm_original.fig_list = self.rm.get_plot_all(y_metric_list=self.vars['y_metrics'],
                                                             x_metric=self.vars['x_metric'],
                                                             groupby_list=self.vars['groupby_list'],
                                                             legend_list=self.vars['legend_list'],
                                                             log_metric_list=self.vars['log_metric_list'],
                                                             mode=self.vars['mode'],
                                                             bar_agg=self.vars['bar_agg'],
                                                             figsize=self.vars['figsize'],
                                                             title_list=self.vars['title_list'],
                                                             legend_format=self.vars['legend_format'],
                                                             title_format=self.vars['title_format'],
                                                             cmap=self.vars['cmap'],
                                                             avg_across=avg_across_value)

            show_inline_matplotlib_plots()

    d_style.observe(on_clicked)
    brefresh.on_click(on_clicked)

    with output:
        display(button)
        display(output_plot)



import cv2 
from . import haven_utils as hu
import numpy as np 
from PIL import Image
from skimage.segmentation import mark_boundaries
    

def mask_on_image(image, mask, add_bbox=False, return_pil=False):
    """[summary]
    
    Parameters
    ----------
    image : [type]
        [description]
    mask : [type]
        [description]
    add_bbox : bool, optional
        [description], by default True
    
    Returns
    -------
    [type]
        [description]
    """
    image = image_as_uint8(image)
    mask = np.array(mask).squeeze()
    obj_ids = np.unique(mask)

    # polygons = cv2.findContours(im,cv2.RETR_TREE,cv2.CHAIN_APPROX_NONE)[1][0]
    red = np.zeros(image.shape, dtype='uint8')
    red[:,:,2] = 255
    alpha = 0.5
    result = image.copy()
    for o in obj_ids:
        if o == 0:
            continue
        ind = mask==o
        result[ind] = result[ind] * alpha + red[ind] * (1-alpha)
        pos = np.where(ind)
        xmin = np.min(pos[1])
        xmax = np.max(pos[1])
        ymin = np.min(pos[0])
        ymax = np.max(pos[0])
        if add_bbox:
            result = cv2.rectangle(result, (xmin, ymin), 
                                                    (xmax, ymax), 
                                                    color=(0,255,0), 
                                                    thickness=2)
    result = mark_boundaries(result, mask) 

    if return_pil:
        return Image.fromarray(result)

    return result

def resize_points(points, h, w):
    """[summary]
    
    Parameters
    ----------
    points : [type]
        [description]
    h : [type]
        [description]
    w : [type]
        [description]
    
    Returns
    -------
    [type]
        [description]
    """
    points = points.squeeze()
    h_old, w_old = points.shape
    y_list, x_list = np.where(points.squeeze())

    points_new = np.zeros((h, w))

    for y, x in zip(y_list, x_list):
        y_new = int((y/h_old) * h)
        x_new = int((x/w_old) * w)
        points_new[y_new, x_new ] = 1

    return points_new

def gray2cmap(gray, cmap="jet", thresh=0):
    """gets a heatmap for a given gray image. Can be used to visualize probabilities.
    
    Parameters
    ----------
    gray : [type]
        [description]
    cmap : str, optional
        [description], by default "jet"
    thresh : int, optional
        [description], by default 0
    
    Returns
    -------
    [type]
        [description]
    """
    # Gray has values between 0 and 255 or 0 and 1
    gray = hu.t2n(gray)
    gray = gray / max(1, gray.max())
    gray = np.maximum(gray - thresh, 0)
    gray = gray / max(1, gray.max())
    gray = gray * 255

    gray = gray.astype(int)
    #print(gray)

    from matplotlib.cm import get_cmap
    cmap = get_cmap(cmap)

    output = np.zeros(gray.shape + (3, ), dtype=np.float64)

    for c in np.unique(gray):
        output[(gray == c).nonzero()] = cmap(c)[:3]

    return hu.l2f(output)


def scatter_plot(X, color, fig=None, title=""):
    """[summary]
    
    Parameters
    ----------
    X : [type]
        [description]
    color : [type]
        [description]
    fig : [type], optional
        [description], by default None
    title : str, optional
        [description], by default ""
    
    Returns
    -------
    [type]
        [description]
    """
    if fig is None:
        fig = plt.figure(figsize=(6,6))
        ax = fig.add_subplot(1, 1, 1) 

    ax = fig.axes[0]

    ax.grid(linestyle='dotted')
    ax.scatter(X[:,0],X[:,1], alpha=0.6, c=color, edgecolors="black")
    

    # plt.axes().set_aspect('equal', 'datalim')
    ax.set_title(title)
    ax.set_xlabel("t-SNE Feature 2")
    ax.set_ylabel("t-SNE Feature 1")
   
    fig.tight_layout(rect=[0, 0.03, 1, 0.95])

    return fig 


def pretty_vis(image, annList, show_class=False, alpha=0.0, dpi=100, **options):
    import cv2
    from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
    from matplotlib.patches import Polygon
    from matplotlib.figure import Figure
    from . import ann_utils as au
    # print(image)
    # if not image.as > 1:
    #     image = image.astype(float)/255.
    image = f2l(image).squeeze().clip(0, 255)
    if image.max() > 1:
        image /= 255.

    # box_alpha = 0.5
    # print(image.clip(0, 255).max())
    color_list = colormap(rgb=True) / 255.

    # fig = Figure()
    fig = plt.figure(frameon=False)
    canvas = FigureCanvas(fig)
    fig.set_size_inches(image.shape[1] / dpi, image.shape[0] / dpi)
    # ax = fig.gca()

    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.axis('off')
    fig.add_axes(ax)
    # im = im.clip(0, 1)
    # print(image)
    ax.imshow(image)

    # Display in largest to smallest order to reduce occlusion
    # areas = (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])
    # sorted_inds = np.argsort(-areas)

    mask_color_id = 0
    for i in range(len(annList)):
        ann = annList[i]

        # bbox = boxes[i, :4]
        # score = boxes[i, -1]

        # bbox = au.ann2bbox(ann)["shape"]
        # score = ann["score"]

        # if score < thresh:
        #     continue

        # show box (off by default, box_alpha=0.0)
        if "bbox" in ann:
            bbox = ann["bbox"]
            ax.add_patch(
                plt.Rectangle((bbox[0], bbox[1]),
                              bbox[2],
                              bbox[3],
                              fill=False,
                              edgecolor='r',
                              linewidth=3.0,
                              alpha=0.5))

        # if show_class:
        # if options.get("show_text") == True or options.get("show_text") is None:
        #     score = ann["score"] or -1
        #     ax.text(
        #         bbox[0], bbox[1] - 2,
        #         "%.1f" % score,
        #         fontsize=14,
        #         family='serif',
        #         bbox=dict(facecolor='g', alpha=1.0, pad=0, edgecolor='none'),
        #         color='white')

        # show mask
        if "segmentation" in ann:
            mask = au.ann2mask(ann)["mask"]
            img = np.ones(image.shape)
            # category_id = ann["category_id"]
            # mask_color_id = category_id - 1
            # color_list = ["r", "g", "b","y", "w","orange","purple"]
            # color_mask = color_list[mask_color_id % len(color_list)]
            color_mask = color_list[mask_color_id % len(color_list), 0:3]
            mask_color_id += 1
            # print("color id: %d - category_id: %d - color mask: %s" 
                        # %(mask_color_id, category_id, str(color_mask)))
            w_ratio = .4
            for c in range(3):
                color_mask[c] = color_mask[c] * (1 - w_ratio) + w_ratio
            for c in range(3):
                img[:, :, c] = color_mask[c]
            e = mask

            contour, hier = cv2.findContours(e.copy(), 
                                    cv2.RETR_CCOMP,
                                    cv2.CHAIN_APPROX_NONE)

            for c in contour:
                polygon = Polygon(
                    c.reshape((-1, 2)),
                    fill=True,
                    facecolor=color_mask,
                    edgecolor="white",
                    linewidth=1.5,
                    alpha=0.7
                    )
                ax.add_patch(polygon)

    canvas.draw()  # draw the canvas, cache the renderer
    width, height = fig.get_size_inches() * fig.get_dpi()
    # image = np.fromstring(canvas.tostring_rgb(), dtype='uint8')

    fig_image = np.fromstring(
        canvas.tostring_rgb(), dtype='uint8').reshape(
            int(height), int(width), 3)
    plt.close()
    # print(fig_image)
    return fig_image



def text_on_image(text, image):
    """Adds test on the image
    
    Parameters
    ----------
    text : [type]
        [description]
    image : [type]
        [description]
    
    Returns
    -------
    [type]
        [description]
    """
    font = cv2.FONT_HERSHEY_SIMPLEX
    bottomLeftCornerOfText = (10,40)
    fontScale              = 0.8
    fontColor              = (1,1,1)
    lineType               = 1
    # img_mask = skimage.transform.rescale(np.array(img_mask), 1.0)
    # img_np = skimage.transform.rescale(np.array(img_points), 1.0)
    img_np = cv2.putText(image, text, 
        bottomLeftCornerOfText, 
        font, 
        fontScale,
        fontColor,
        thickness=2
        # lineType
        )
    return img_np

def bbox_on_image(bbox_xyxy, image, mode='yxyx', color=(255, 0, 0)):
    """[summary]
    
    Parameters
    ----------
    bbox_xyxy : [type]
        [description]
    image : [type]
        [description]
    mode : str, optional
        [description], by default 'xyxy'
    
    Returns
    -------
    [type]
        [description]
    """
    image_uint8 = image_as_uint8(image)

    H, W, _ = image_uint8.shape

    for bb in bbox_xyxy:
        if mode == 'yxyx':
            y1, x1, y2, x2 = bb
        else:
            x1, y1, x2, y2 = bb

        if mode == 'xywh':
            x2 += x1
            y2 += y1

        if x2 < 1:
            start_point = (int(x1*W), int(y1*H), ) 
            end_point = ( int(x2*W), int(y2*H),)
        else:
            start_point = (int(x1), int(y1))
            end_point = (int(x2), int(y2))
        
        # Line thickness of 2 px 
        thickness = 2
        # Draw a rectangle with blue line borders of thickness of 2 px 
        image_uint8 = cv2.rectangle(image_uint8.copy(), start_point, end_point, color, thickness) 

    return image_uint8 / 255.

def points_on_image(y_list, x_list, image, radius=3, c_list=None):
    """[summary]
    
    Parameters
    ----------
    y_list : [type]
        [description]
    x_list : [type]
        [description]
    image : [type]
        [description]
    radius : int, optional
        [description], by default 3
    
    Returns
    -------
    [type]
        [description]
    """
    image_uint8 = image_as_uint8(image)

    H, W, _ = image_uint8.shape
    color_list = [(255, 0, 0) , (0, 255, 0) , (0, 0, 255) ]
    for i, (y, x) in enumerate(zip(y_list, x_list)):
        if y < 1:
            x, y = int(x*W), int(y*H) 
        else:
            x, y = int(x), int(y) 
            
        # Blue color in BGR 
        if c_list is not None:
            color = color_list[c_list[i]] 
        else:
            color = color_list[1] 
        
        # Line thickness of 2 px 
        thickness = 5
        # Using cv2.rectangle() method 
        # Draw a rectangle with blue line borders of thickness of 2 px 
        image_uint8 = cv2.circle(image_uint8, (x,y), radius, color, thickness) 

        start_point = (x-radius*2, y-radius*2) 
        end_point = (x+radius*2, y+radius*2) 
        thickness = 2
        color = (255, 0, 0)
        
        image_uint8 = cv2.rectangle(image_uint8, start_point, end_point, color, thickness) 

    return image_uint8 / 255.


def image_as_uint8(img):
    """Returns a uint8 version of the image
    
    Parameters
    ----------
    img : [type]
        [description]
    
    Returns
    -------
    [type]
        [description]
    """
    image = hu.f2l(np.array(img).squeeze())
    
    if image.dtype != 'uint8':
        image_uint8 = (image*255).astype("uint8").copy()
    else:
        image_uint8 = image 

    return image_uint8



from .. import haven_utils
from .. import haven_results as hr
from .. import haven_utils as hu
from .. import haven_share as hd

from .share_tab import share_tab
from .plots_tab import plots_tab
from .tables_tab import tables_tab
from .latex_tab import latex_tab
from .images_tab import images_tab
from . import utils_widgets as uw

import os
import pprint
import json
import copy
import pprint
import pandas as pd

try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except:
    print('widgets not available...')


def get_dashboard(rm, vars=None, show_jobs=True, wide_display=False, enable_datatables=True):
    dm = DashboardManager(
        rm, vars=vars, show_jobs=show_jobs, wide_display=wide_display, enable_datatables=enable_datatables)
    dm.display()
    return dm
    
class DashboardManager:
    def __init__(self, rm, vars=None, show_jobs=True, wide_display=True, enable_datatables=True):
        self.rm_original = rm
        if vars is None:
            self.vars = {}
        else:
            self.vars = vars

        self.show_jobs = show_jobs
        self.wide_display = wide_display
        self.enable_datatables = enable_datatables

        self.layout = widgets.Layout(width='100px')
        self.layout_label = widgets.Layout(width='200px')
        self.layout_dropdown = widgets.Layout(width='200px')
        self.layout_button = widgets.Layout(width='200px')
        self.t_savedir_base = widgets.Text(
            value=str(self.vars.get('savedir_base') or rm.savedir_base),
            layout=widgets.Layout(width='600px'),
            disabled=False
        )

        self.t_filterby_list = widgets.Text(
            value=str(self.vars.get('filterby_list')),
            layout=widgets.Layout(width='1200px'),
            description='               filterby_list:',
            disabled=False
        )

    def display(self):
        self.update_rm()

        # Select Exp Group
        l_exp_group = widgets.Label(
            value="Select exp_group", layout=self.layout_label,)

        exp_group_list = list(self.rm_original.exp_groups.keys())
        exp_group_selected = 'all'
        if self.vars.get('exp_group', 'all') in exp_group_list:
            exp_group_selected = self.vars.get('exp_group', 'all')

        d_exp_group = widgets.Dropdown(
            options=exp_group_list,
            value=exp_group_selected,
            layout=self.layout_dropdown,
        )
        self.rm_original.exp_list_all = self.rm_original.exp_groups.get(
            d_exp_group.value, 'all')
        l_n_exps = widgets.Label(value='Total Exps %d' % len(
            self.rm_original.exp_list_all), layout=self.layout,)

        def on_group_change(change):
            if change['type'] == 'change' and change['name'] == 'value':
                self.rm_original.exp_list_all = self.rm_original.exp_groups[change['new']]
                l_n_exps.value = 'Total Exps %d' % len(
                    self.rm_original.exp_list_all)

        d_exp_group.observe(on_group_change)

        display(widgets.VBox([l_exp_group,
                              widgets.HBox(
                                  [d_exp_group, l_n_exps, self.t_filterby_list])
                              ]))
        
        if self.enable_datatables:
            init_datatable_mode()
        tables = widgets.Output()
        plots = widgets.Output()
        images = widgets.Output()
        share = widgets.Output()
        latex = widgets.Output()

        main_out = widgets.Output()
        # Display tabs
        tab = widgets.Tab(children=[tables, plots, images, share, latex])
        tab.set_title(0, 'Tables')
        tab.set_title(1, 'Plots')
        tab.set_title(2, 'Images')
        tab.set_title(3, 'Share Results')
        tab.set_title(4, 'Latex')

        with main_out:
            display(tab)
            tables.clear_output()
            plots.clear_output()
            images.clear_output()
            share.clear_output()
            latex.clear_output()

            # show tabs
            tables_tab(self, tables)
            plots_tab(self, plots)
            images_tab(self, images)
            share_tab(self, share)
            latex_tab(self, latex)

        display(main_out)

        if self.wide_display:
            display(
                HTML("<style>.container { width:100% !important; }</style>"))

        # This makes cell show full height display
        style = """
        <style>
            .output_scroll {
                height: unset !important;
                border-radius: unset !important;
                -webkit-box-shadow: unset !important;
                box-shadow: unset !important;
            }
        </style>
        """
        display(HTML(style))

    def update_rm(self):
        self.rm = hr.ResultManager(exp_list=self.rm_original.exp_list_all,
                                   savedir_base=str(self.t_savedir_base.value),
                                   filterby_list=hu.get_dict_from_str(
                                       str(self.t_filterby_list.value)),
                                   verbose=self.rm_original.verbose,
                                   mode_key=self.rm_original.mode_key,
                                   has_score_list=self.rm_original.has_score_list,
                                   score_list_name=self.rm_original.score_list_name
                                   )

        if len(self.rm.exp_list) == 0:
            if self.rm.n_exp_all > 0:
                display('No experiments selected out of %d '
                        'for filtrby_list %s' % (self.rm.n_exp_all,
                                                 self.rm.filterby_list))
                display('Table below shows all experiments.')
                score_table = hr.get_score_df(exp_list=self.rm_original.exp_list_all,
                                              savedir_base=self.rm_original.savedir_base)
                display(score_table)
            else:
                display('No experiments exist...')
            return
        else:
            display('Selected %d/%d experiments using "filterby_list" from savedir_base "%s"' %
                    (len(self.rm.exp_list), len(self.rm.exp_list_all), self.vars.get('savedir_base')))



def launch_jupyter():
    """
    virtualenv -p python3 .
    source bin/activate
    pip install jupyter notebook
    jupyter notebook --ip 0.0.0.0 --port 2222 --NotebookApp.token='abcdefg'
    """
    print()


def create_jupyter(fname='example.ipynb',
                   savedir_base='<path_to_saved_experiments>',
                   overwrite=False, print_url=False,
                   create_notebook=True):
    print('Jupyter')

    if create_notebook and (overwrite or not os.path.exists(fname)):
        cells = [main_cell(savedir_base), install_cell()]
        os.makedirs(os.path.dirname(fname), exist_ok=True)
        save_ipynb(fname, cells)
        print('- saved:', fname)

    if print_url:
        from notebook import notebookapp
        servers = list(notebookapp.list_running_servers())
        hostname = os.uname().nodename

        flag = False
        for i, s in enumerate(servers):
            if s['hostname'] == 'localhost':
                continue
            flag = True
            url = 'http://%s:%s/' % (hostname, s['port'])
            print('- url:', url)

        if flag == False:
            print('a jupyter server was not found :(')
            print(
                'a jupyter server can be started using the script in https://github.com/ElementAI/haven .')


def main_cell(savedir_base):
    script = ("""
from haven import haven_jupyter as hj
from haven import haven_results as hr
from haven import haven_utils as hu

# path to where the experiments got saved
savedir_base = '%s'
exp_list = None

# filter exps
# e.g. filterby_list =[{'dataset':'mnist'}] gets exps with mnist
filterby_list = None

# get experiments
rm = hr.ResultManager(exp_list=exp_list, 
                      savedir_base=savedir_base, 
                      filterby_list=filterby_list,
                      verbose=0,
                      exp_groups=None
                     )

# launch dashboard
# make sure you have 'widgetsnbextension' enabled; 
# otherwise see README.md in https://github.com/ElementAI/haven

hj.get_dashboard(rm, vars(), wide_display=True)
          """ % savedir_base)
    return script


def install_cell():
    script = ("""
    !pip install --upgrade git+https://github.com/ElementAI/haven
          """)
    return script


def save_ipynb(fname, script_list):
    import nbformat as nbf

    nb = nbf.v4.new_notebook()
    nb['cells'] = [nbf.v4.new_code_cell(code) for code in
                   script_list]
    with open(fname, 'w') as f:
        nbf.write(nb, f)


def init_datatable_mode():
    """Initialize DataTable mode for pandas DataFrame represenation."""
    import pandas as pd
    from IPython.core.display import display, Javascript

    # configure path to the datatables library using requireJS
    # that way the library will become globally available
    display(Javascript("""
        require.config({
            paths: {
                DT: '//cdn.datatables.net/1.10.19/js/jquery.dataTables.min',
            }
        });
        $('head').append('<link rel="stylesheet" type="text/css" href="//cdn.datatables.net/1.10.19/css/jquery.dataTables.min.css">');
    """))

    def _repr_datatable_(self):
        """Return DataTable representation of pandas DataFrame."""
        # classes for dataframe table (optional)
        classes = ['table', 'table-striped', 'table-bordered']

        # create table DOM
        script = (
            f'$(element).html(`{self.to_html(index=True, classes=classes)}`);\n'
        )

        # execute jQuery to turn table into DataTable
        script += """
            require(["DT"], function(DT) {
                $(document).ready( () => {
                    // Turn existing table into datatable
                    $(element).find("table.dataframe").DataTable({"scrollX": true});

                    $('#container').css( 'display', 'block' );
                    table.columns.adjust().draw();
                    
                })
            });
        """

        return script

    pd.DataFrame._repr_javascript_ = _repr_datatable_






import ast


def get_dict_from_str(string):
    if string is None:
        return string

    if string == 'None':
        return None

    if string == '':
        return None
        
    return ast.literal_eval(string)

def get_list_from_str(string):
    if string is None:
        return string

    if string == 'None':
        return None

    string = string.replace(' ','').replace(']', '').replace('[', '').replace('"', '').replace("'", "")

    if string == '':
        return None

    return string.split(',')



from .. import haven_utils
from .. import haven_results as hr
from .. import haven_utils as hu 
from .. import haven_share as hd

import os
import pprint, json
import copy
import pprint
import pandas as pd 

try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except:
    print('widgets not available...')

def latex_tab(self, output):
    kwargs = {}

    filter_dict = self.vars.get('filter_dict', {})
    map_row_dict_dict = self.vars.get('map_row_dict_dict', {})
    map_col_dict = self.vars.get('map_col_dict', {})
    table = (self.rm.get_score_table())
    # map columns
    table2 = pd.DataFrame()
    for col_old, col_new in map_col_dict.items():
        # map column
        table2[col_new] = table[col_old]
        
        # map rows
        if col_old in map_row_dict_dict:
            map_row_dict = map_row_dict_dict[col_old]
            table2[col_new] = table2[col_new].apply( lambda x: map_row_dict[x.replace("'","")] if x.replace("'","") in map_row_dict else x )
        
    # filter dict
    conds = None
    for k, v in filter_dict.items():
        if not isinstance(v, list):
            v = [v]
#         print(k, v)
        for vi in v:
            cond = table2[k] == vi
            if conds is None:
                conds = cond
            else:
                conds = conds | cond
        
        table2 = table2[conds]
        table2 = table2.set_index(k)
        print(v)
        table2 = table2.reindex(v)
        table2.insert(0, k, table2.index)
        table2 = table2.reset_index(drop=True)
        
    with output:
        display(table2.to_latex(**kwargs))

def create_latex_table(table, filter_dict, map_row_dict_dict, map_col_dict, **kwargs):
    '''
    Usage
    -----
        map_row_dict_dict = {'model.loss': {"point_loss":'Point Loss', 
                                        "cons_point_loss":'CB Point Loss', 
                                        'joint_cross_entropy':'W-CE (Full Sup.)'}}

        map_col_dict = {'model.loss': 'Loss Function', 
                        'test_dice':'Dice', 
                        'test_iou':'IoU', 
                        'test_prec':'PPV', 
                        'test_recall':'Sens.', 
                        'test_spec':'Spec.'}

        filter_dict = {'Loss Function':['Point Loss', 'CB Point Loss','W-CE (Full Sup.)']}

        caption_dict = {'weakly_covid19_v1_c2':'COVID-19-A',
                        'weakly_covid19_v2_mixed_c2':'COVID-19-B-Mixed',
                        'weakly_covid19_v3_mixed_c2':'COVID-19-C-Mixed',
                        
                        'weakly_covid19_v2_sep_c2':'COVID-19-B-Sep',
                        'weakly_covid19_v3_sep_c2':'COVID-19-C-Sep'}

        for exp_name in ['weakly_covid19_v1_c2', 
                        
                        'weakly_covid19_v2_mixed_c2', 
                        'weakly_covid19_v3_mixed_c2',
                        
                        'weakly_covid19_v2_sep_c2', 
                        'weakly_covid19_v3_sep_c2']:
            rm.exp_list = hr.get_exp_list_from_config([exp_name], exp_config_name)
            table = (rm.get_score_table())
            print(create_latex_table(table=table, 
                                    filter_dict=filter_dict, 
                                    map_row_dict_dict=map_row_dict_dict, 
                                    map_col_dict=map_col_dict,
                                    float_format='%.2f', 
                                    caption=caption_dict[exp_name], 
        #                              label=caption_dict, 
                                    index=False))
                                    
    '''
    pass


import argparse, pandas as pd


def joke():
    sdfds
    parser = argparse.ArgumentParser()

    parser.add_argument('-e', '--exp_group_list')
    parser.add_argument('-sb', '--savedir_base', required=True)
    parser.add_argument('-ei', '--exp_id', default=None)

    args = parser.parse_args()
    print(args)


def run_file():
    print('lol')


import os 
import copy
import hashlib
import itertools
from itertools import groupby
import tqdm, pprint
import numpy as np
from .. import haven_utils as hu

def cartesian_exp_group(exp_config, remove_none):
    """Cartesian experiment config.

    It converts the exp_config into a list of experiment configuration by doing
    the cartesian product of the different configuration. It is equivalent to
    do a grid search.

    Parameters
    ----------
    exp_config : str
        Dictionary with the experiment Configuration

    Returns
    -------
    exp_list: str
        A list of experiments, each defines a single set of hyper-parameters
    """
    exp_config_copy = copy.deepcopy(exp_config)

    # Make sure each value is a list
    for k, v in exp_config_copy.items():
        if not isinstance(exp_config_copy[k], list):
            exp_config_copy[k] = [v]

    # Create the cartesian product
    exp_list_raw = (dict(zip(exp_config_copy.keys(), values))
                    for values in itertools.product(*exp_config_copy.values()))

    # Convert into a list
    exp_list = []
    for exp_dict in exp_list_raw:
        # remove hparams with None
        if remove_none:
            to_remove = []
            for k, v in exp_dict.items():
                if v is None:
                    to_remove += [k]
            for k in to_remove:
                del exp_dict[k]
        exp_list += [exp_dict]

    return exp_list


def hash_dict(exp_dict):
    """Create a hash for an experiment.

    Parameters
    ----------
    exp_dict : dict
        An experiment, which is a single set of hyper-parameters

    Returns
    -------
    hash_id: str
        A unique id defining the experiment
    """
    dict2hash = ""
    if not isinstance(exp_dict, dict):
        raise ValueError('exp_dict is not a dict')

    for k in sorted(exp_dict.keys()):
        if '.' in k:
            raise ValueError(". has special purpose")
        elif isinstance(exp_dict[k], dict):
            v = hash_dict(exp_dict[k])
        elif isinstance(exp_dict[k], tuple):
            raise ValueError("tuples can't be hashed yet, consider converting tuples to lists")
        else:
            v = exp_dict[k]

        dict2hash += os.path.join(str(k), str(v))

    hash_id = hashlib.md5(dict2hash.encode()).hexdigest()

    return hash_id

def hash_str(str):
    """Hashes a string using md5.

    Parameters
    ----------
    str
        a string

    Returns
    -------
    str
        md5 hash for the input string
    """
    return hashlib.md5(str.encode()).hexdigest()



def group_exp_list(exp_list, groupby_list):
    """Split the experiment list into smaller lists where each
       is grouped by a set of hyper-parameters

    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    groupby_list : list
        The set of hyper-parameters to group the experiments

    Returns
    -------
    lists_of_exp_list : list
        Experiments grouped by a set of hyper-parameters

    Example
    -------
    >>>
    >>>
    >>>
    """
    if groupby_list is None:
        return [exp_list]
    if not isinstance(groupby_list, list):
        groupby_list = [groupby_list]
    # groupby_list = hu.as_double_list(groupby_list)
    def split_func(x):
        x_list = []
        for k_list in groupby_list:
            if not isinstance(k_list, list):
                k_list = [k_list]
            val = get_str(x, k_list)
            x_list += [val]

        return x_list

    exp_list.sort(key=split_func)

    list_of_exp_list = []
    group_dict = groupby(exp_list, key=split_func)

    # exp_group_dict = {}
    for k, v in group_dict:
        v_list = list(v)
        list_of_exp_list += [v_list]
    #     # print(k)
    #     exp_group_dict['_'.join(list(map(str, k)))] = v_list

    return list_of_exp_list

def group_list(python_list, key, return_count=False):
    group_dict = {}
    for p in python_list:
        p_tmp = copy.deepcopy(p)
        del p_tmp[key]
        k = p[key]
        
        if k not in group_dict:
            group_dict[k] = []

        group_dict[k] += [p_tmp]
    
    if return_count:
        count_dict = {}
        for k in group_dict:
            count_dict[k] = len(group_dict[k])
        return count_dict
    return group_dict

def get_exp_list_from_config(exp_groups, exp_config_fname):
    exp_list = []
    for e in exp_groups:
        exp_list += hu.load_py(exp_config_fname).EXP_GROUPS[e]

    return exp_list

def get_str(h_dict, k_list):
    k = k_list[0]

    if len(k_list) == 1:
        return str(h_dict.get(k))
    
    return get_str(h_dict.get(k), k_list[1:])

def get_best_exp_dict(exp_list, savedir_base, metric, 
                      metric_agg='min', filterby_list=None, 
                      avg_across=None,
                      return_scores=False, verbose=True,
                        score_list_name='score_list.pkl'):
    """Obtain best the experiment for a specific metric.

    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    savedir_base : [type]
        A directory where experiments are saved
    metric : [type]
        [description]
    min_or_max : [type]
        [description]
    return_scores : bool, optional
        [description], by default False
    """
    scores_dict = []
    if metric_agg in ['min', 'min_last']:
        best_score = np.inf
    elif metric_agg in ['max', 'max_last']:
        best_score = 0.
    
    exp_dict_best = None
    exp_list = filter_exp_list(exp_list, filterby_list, verbose=verbose)
    for exp_dict in exp_list:
        exp_id = hu.hash_dict(exp_dict)
        savedir = os.path.join(savedir_base, exp_id)

        score_list_fname = os.path.join(savedir, score_list_name)
        if not os.path.exists(score_list_fname):
            if verbose:
                print('%s: missing %s' % (exp_id, score_list_name))
            continue
        
        score_list = hu.load_pkl(score_list_fname)

        if metric_agg in ['min', 'min_last']:
            if metric_agg == 'min_last':
                score = [score_dict[metric] for score_dict in score_list][-1]
            elif metric_agg == 'min':
                score = np.min([score_dict[metric] for score_dict in score_list])
            if best_score >= score:
                best_score = score
                exp_dict_best = exp_dict

        elif metric_agg in ['max', 'max_last']:
            if metric_agg == 'max_last':
                score = [score_dict[metric] for score_dict in score_list][-1]
            elif metric_agg == 'max':
                score = np.max([score_dict[metric] for score_dict in score_list])
                
            if best_score <= score:
                best_score = score
                exp_dict_best = exp_dict

        scores_dict += [{'score': score,
                         'epochs': len(score_list), 
                         'exp_id': exp_id}]

    if exp_dict_best is None:
        if verbose:
            print('no experiments with metric "%s"' % metric)
        return {}
        
    return exp_dict_best


def get_exp_list_from_exp_configs(exp_group_list, workdir, filterby_list=None, verbose=True):
    """[summary]
    
    Parameters
    ----------
    exp_group_list : [type]
        [description]
    workdir : [type]
        [description]
    filterby_list : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    assert workdir is not None
    
    from importlib import reload
    assert(workdir is not None)
    if workdir not in sys.path:
        sys.path.append(workdir)
    import exp_configs as ec
    reload(ec)

    exp_list = []
    for exp_group in exp_group_list:
        exp_list += ec.EXP_GROUPS[exp_group]
    if verbose:
        print('%d experiments' % len(exp_list))

    exp_list = filter_exp_list(exp_list, filterby_list, verbose=verbose)
    return exp_list

def get_exp_list(savedir_base, filterby_list=None, verbose=True):
    """[summary]
    
    Parameters
    ----------
    savedir_base : [type], optional
        A directory where experiments are saved, by default None
    filterby_list : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    exp_list = []
    dir_list = os.listdir(savedir_base)

    for exp_id in tqdm.tqdm(dir_list):
        savedir = os.path.join(savedir_base, exp_id)
        fname = os.path.join(savedir, 'exp_dict.json')
        if len(exp_id) != 32:
            if verbose:
                print('"%s/" is not an exp directory' % exp_id)
            continue 

        if not os.path.exists(fname):
            if verbose:
                print('%s: missing exp_dict.json' % exp_id)
            continue
        
        exp_dict = hu.load_json(fname)
        expected_id = hu.hash_dict(exp_dict)
        if expected_id != exp_id:
            if verbose:
                # assert(hu.hash_dict(exp_dict) == exp_id)
                print('%s does not match %s' % (expected_id, exp_id))
            continue
        # print(hu.hash_dict(exp_dict))
        exp_list += [exp_dict]

    exp_list = filter_exp_list(exp_list, filterby_list)
    return exp_list


def zip_exp_list(savedir_base):
    """[summary]

    Parameters
    ----------
    savedir_base : [type]
        [description]
    """
    import zipfile

    with zipfile.ZipFile(savedir_base) as z:
        for filename in z.namelist():
            if not os.path.isdir(filename):
                # read the file
                with z.open(filename) as f:
                    for line in f:
                        print(line)

def filter_list(python_list, filterby_list, verbose=True):
    return filter_exp_list(python_list, filterby_list, verbose=verbose)

def filter_exp_list(exp_list, filterby_list, savedir_base=None, verbose=True,
                    score_list_name='score_list.pkl', return_style_list=False):
    """[summary]
    
    Parameters
    ----------
    exp_list : [type]
        A list of experiments, each defines a single set of hyper-parameters
    filterby_list : [type]
        [description]
    
    Returns
    -------
    [type]
        [description]
    """
    if filterby_list is None or filterby_list == '' or len(filterby_list) == 0:
        if return_style_list:
            return exp_list, [{}]*len(exp_list)
        else:
            return exp_list

    style_list = []
    filterby_list_list = hu.as_double_list(filterby_list)
    # filterby_list = filterby_list_list
    
    for filterby_list in filterby_list_list:
        exp_list_new = []

        # those with meta
        filterby_list_no_best = []
        for filterby_dict in filterby_list:
            meta_dict = {}
            if isinstance(filterby_dict, tuple):
                fd, meta_dict = filterby_dict
            
            if meta_dict.get('best'):
                assert savedir_base is not None
                el = filter_exp_list(exp_list, filterby_list=fd, verbose=verbose)
                best_dict = meta_dict.get('best')
                exp_dict = get_best_exp_dict(el, savedir_base, 
                                metric=best_dict['metric'],
                                metric_agg=best_dict['metric_agg'], 
                                filterby_list=None, 
                                avg_across=best_dict.get('avg_across'),
                                return_scores=False, 
                                verbose=verbose,
                                score_list_name=score_list_name)

                exp_list_new += [exp_dict]
                style_list += [meta_dict.get('style', {})]
            else:
                filterby_list_no_best += [filterby_dict] 

        
        # ignore metas here meta
        for exp_dict in exp_list:
            select_flag = False
            
            for fd in filterby_list_no_best:
                if isinstance(fd, tuple):
                    filterby_dict, meta_dict = fd
                    style_dict = meta_dict.get('style', {})
                else:
                    filterby_dict = fd
                    style_dict = {}

                filterby_dict = copy.deepcopy(filterby_dict)
               
                keys = filterby_dict.keys()
                for k in keys:
                    if '.' in k:
                        v = filterby_dict[k]
                        k_list = k.split('.')
                        nk = len(k_list)

                        dict_tree = dict()
                        t = dict_tree

                        for i in range(nk):
                            ki = k_list[i]
                            if i == (nk - 1):
                                t = t.setdefault(ki, v)
                            else:
                                t = t.setdefault(ki, {})

                        filterby_dict = dict_tree

                assert (isinstance(filterby_dict, dict), 
                                 ('filterby_dict: %s is not a dict' % str(filterby_dict)))

                if hu.is_subset(filterby_dict, exp_dict):
                    select_flag = True
                    break

            if select_flag:
                exp_list_new += [exp_dict]
                style_list += [style_dict]

        exp_list = exp_list_new
        

    if verbose:
        print('Filtered: %d/%d experiments gathered...' % (len(exp_list_new), len(exp_list)))
    # hu.check_duplicates(exp_list_new)
    exp_list_new = hu.ignore_duplicates(exp_list_new)
    
    if return_style_list:
        return exp_list_new, style_list

    return exp_list_new





import torch, torchvision, os, pprint
import tqdm
import argparse, pandas as pd

from haven import haven_utils as hu
from haven import haven_wizard as hw

def get_loader(name, split, datadir, exp_dict):
    if name == 'mnist':
        # get dataset and loader
        train = True if split == 'train' else False
        transform = torchvision.transforms.Compose([torchvision.transforms.ToTensor()])
        dataset = torchvision.datasets.MNIST(root=datadir, 
                                             train=train, download=True,
                                             transform=transform)
        loader = torch.utils.data.DataLoader(dataset, batch_size=256)

    return loader

def get_model(name, exp_dict):
    if name == 'linear':
        model = Linear()
        model.opt = torch.optim.Adam(model.parameters(), lr=exp_dict['lr'])

    if name == 'mlp':
        model = Mlp()
        model.opt = torch.optim.Adam(model.parameters(), lr=exp_dict['lr'])

    return model

# =====================================================
class Linear(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.model = torch.nn.Linear(784, 10)

    def train_on_batch(self, batch, **extras):
        self.opt.zero_grad()
        self.train()

        images, labels = batch
        logits = self.model.forward(images.view(images.shape[0], -1))
        criterion = torch.nn.CrossEntropyLoss(reduction="mean")
        loss = criterion(logits, labels.view(-1))
        loss.backward()

        self.opt.step()

        return {"train_loss": loss.item(), 'train_acc':(logits.argmax(dim=1) == labels).float().mean().item()}


# MLP
class Mlp(torch.nn.Module):
    def __init__(self, input_size=784,
                 hidden_sizes=[512, 256],
                 n_classes=10,
                 bias=True, dropout=False):
        super().__init__()
        self.input_size = input_size
        self.hidden_layers = torch.nn.ModuleList([torch.nn.Linear(in_size, out_size, bias=bias) for
                                            in_size, out_size in zip([self.input_size] + hidden_sizes[:-1], hidden_sizes)])
        self.output_layer = torch.nn.Linear(hidden_sizes[-1], n_classes, bias=bias)

    def train_on_batch(self, batch, **extras):
        self.opt.zero_grad()
        self.train()

        images, labels = batch
        logits = self.forward(images)
        criterion = torch.nn.CrossEntropyLoss(reduction="mean")
        loss = criterion(logits, labels.view(-1))
        loss.backward()

        self.opt.step()

        return {"train_loss": loss.item(), 'train_acc':(logits.argmax(dim=1) == labels).float().mean().item()}

    def forward(self, x):
        x = x.view(-1, self.input_size)
        out = x
        for layer in self.hidden_layers:
            Z = layer(out)
            out = torch.nn.functional.relu(Z)

        logits = self.output_layer(out)

        return logits

    def get_state_dict(self):
        state_dict = {"model": self.state_dict(),
                      "opt":self.opt.state_dict()}

        return state_dict
        
    def set_state_dict(self, state_dict):
        self.load_state_dict(state_dict["model"])
        self.opt.load_state_dict(state_dict["opt"])




    


import copy
import glob
import os
import sys
import pprint
from itertools import groupby
from textwrap import wrap
import numpy as np
import pandas as pd
import pylab as plt
import tqdm

def get_bar_chart(score_list, label_list, sep, ylabel, fontsize, title, width, legend_flag=False, figsize=(20,10)):
    """
    label_list = ['LCFCN', 'WSLM', 'Glance-ram', 'Glance']
    score_list = [2.12, 3.10, 3.38, 4.03]
    get_bar_chart(score_list=score_list, label_list=label_list, ylabel='MAE', width=0.35, sep=.5, fontsize=fontsize,
                                    title='Mall')
    """
    fig, ax = plt.subplots(figsize = figsize, dpi=200)
    ind = np.arange(len(score_list))*sep
    plt.title(title, fontsize=fontsize+4)
    width = width   
    for i in range(len(ind)):
        index = ind[i]
        value = score_list[i]

        rects = ax.bar([index+width], [value], width=width, label=label_list[i])
    
    minimum, maximum = ax.get_ylim()
    y = .05 * (maximum - minimum)

    for i in range(len(ind)):
        index = ind[i]
        value = score_list[i]
    
        ax.text(x=index + width - .025 , y = y, s=f"{value}" , fontdict=dict(fontsize=(fontsize or 10)+2), color='white')

    ax.tick_params(axis='x', labelsize=fontsize)
    ax.tick_params(axis='y', labelsize=fontsize)

    ax.set_ylabel(ylabel, fontsize=fontsize+2)
    ax.grid(True)

    if legend_flag:
        plt.legend(**{'fontsize':fontsize, "loc":2, "bbox_to_anchor":(1.05,1),
                                              'borderaxespad':0., "ncol":1})
    else:
        ax.set_xticks(ind + width)
        ax.set_xticklabels(label_list)
    plt.tight_layout()
    
    plt.show()
    



import numpy as np
import copy
import pycocotools.mask as mask_util
from itertools import product 
import torch


class APMonitor:
    def __init__(self):
        self.pred_ann_list = []
        self.gt_ann_list = []
        self.n_batches = 0.
        self.iou_thr = 0.5
        self.iou_thr_list = [0.5, 0.75]

    def add(self, gt_ann_list, pred_ann_list):
        self.pred_ann_list += pred_ann_list
        self.gt_ann_list += gt_ann_list

        self.n_batches += 1.

    def get_avg_score(self):
        results = compute_precision(self.gt_ann_list, 
                                         self.pred_ann_list,
                                         iouType='segm',
                                         iouThr=0.5,
                                         iouThrList=np.array([0.25,0.5,0.75]))
        results = {k:v for k,v in results.items() if 'mAP' in k}
        results["val_score"] = results['mAP50.0']


        return results

def compute_precision(gt_annList, pred_annList, iouType, iouThr, iouThrList):
    gt_annList = copy.deepcopy(gt_annList)
    # pred_annList = copy.deepcopy(pred_annList)
    if len(gt_annList) == 0:
        return {'mAP50.0': 0}
    if len(pred_annList) == 0:
        return {'mAP50.0': 0}

    result_dict = evaluate_annList(pred_annList=pred_annList,
                                                gt_annList=gt_annList,
                                                iouType=iouType,
                                                iouThr=iouThr,
                                                ap=1,
                                                iouThrList=iouThrList)

    return result_dict


def evaluate_annList(pred_annList, gt_annList,
                     ap=1, iouType="bbox",
                     iouThr=0.5,
                     maxDets=100,
                     aRngLabel="all",
                     iouThrList=None):
    if iouThrList is None:
        iouThrList = np.linspace(.5, 0.95, int(np.round((0.95 - .5) / .05)) + 1, endpoint=True)

    # print("iouThrList", iouThrList)
    recThrList = np.linspace(.0, 1.00, int(np.round((1.00 - .0) / .01)) + 1, endpoint=True)

    areaRngList = [[0 ** 2, 1e5 ** 2],
                   [0 ** 2, 32 ** 2],
                   [32 ** 2, 96 ** 2],
                   [96 ** 2, 1e5 ** 2]]
    aRngLabelList = ['all', 'small', 'medium', 'large']
    maxDetList = [1, 10, 100]

    # Put them in dict
    gt_dict = {}  # gt for evaluation
    dt_dict = {}  # dt for evaluation
    imgList = set()
    catList = set()

    for i, gt in enumerate(copy.deepcopy(gt_annList)):
        key = (gt['image_id'], gt['category_id'])
        if key not in gt_dict:
            gt_dict[key] = []

        gt["id"] = i + 1
        gt_dict[key] += [gt]

        imgList.add(gt['image_id'])
        catList.add(gt['category_id'])

    for i, dt in enumerate(copy.deepcopy(pred_annList)):
        key = (dt['image_id'], dt['category_id'])
        if key not in dt_dict:
            dt_dict[key] = []

        bb = dt["bbox"]
        dt['area'] = bb[2] * bb[3]

        dt["id"] = i + 1
        dt['iscrowd'] = 0

        dt_dict[key] += [dt]

        # imgList.add(dt['image_id'])
        # catList.add(dt['category_id'])

    imgList = list(imgList)
    catList = list(catList)

    # compute ious
    iou_dict = {}
    for imgId, catId in product(imgList, catList):
        key = (imgId, catId)

        if key in gt_dict:
            gt = gt_dict[key]
        else:
            gt = []

        if key in dt_dict:
            dt = dt_dict[key]
        else:
            dt = []

        iou_dict[key] = computeIoU(gt, dt,
                                   iouType=iouType,
                                   maxDets=maxDets)
    # evaluate detections
    evalImgs = []
    for catId, areaRng, imgId in product(catList,
                                         areaRngList,
                                         imgList):
        evalImgs += [
            evaluateImg(gt_dict, dt_dict, iou_dict,
                        imgId, catId, areaRng,
                        maxDets, iouThrs=iouThrList)
        ]

    eval_dict = accumulate(evalImgs, imgList,
                           catList, iouThrList,
                           recThrList,
                           areaRngList, maxDetList)

    result_dict = {"iouThrList": iouThrList,
                   "iouType": iouType,
                   "n_classes": len(catList),
                   "categories": catList}

    # Get thresholds for individual APs
    aind = [i for i, aRng in enumerate(aRngLabelList) if aRng == aRngLabel]
    mind = [i for i, mDet in enumerate(maxDetList) if mDet == maxDets]

    for thr in iouThrList:
        ap_dict = compute_AP_dict(ap, aind, mind,
                                  eval_dict, thr,
                                  iouThrList)
        result_dict.update(ap_dict)

    if ap == 1:
        result_dict["mAP"] = result_dict["mAP%s" % (iouThr * 100)]
    else:
        result_dict["mRC"] = result_dict["mRC%s" % (iouThr * 100)]
    return result_dict


def computeIoU(gt, dt, iouType="bbox", maxDets=100):
    if len(gt) == 0 and len(dt) == 0:
        return []

    inds = np.argsort([-d['score'] for d in dt], kind='mergesort')
    dt = [dt[i] for i in inds]
    if len(dt) > maxDets:
        dt = dt[0:maxDets]

    if iouType == 'segm':
        g = [g['segmentation'] for g in gt]
        d = [d['segmentation'] for d in dt]
    elif iouType == 'bbox':
        g = [g['bbox'] for g in gt]
        d = [d['bbox'] for d in dt]
    else:
        raise Exception('unknown iouType for iou computation')

    # compute iou between each dt and gt region
    iscrowd = [int(o['iscrowd']) for o in gt]

    ious = mask_util.iou(d, g, iscrowd)

    # ####
    # print("id:%s - cat:%d" % (imgId, catId))
    # print("d:", d)
    # print("g:", g)
    # print("ious", ious)
    # ####
    return ious

def compute_AP_dict(ap, aind, mind, eval_dict, iouThr, iouThrList):
    if ap == 1:
        # dimension of precision: [TxRxKxAxM]
        s = eval_dict['precision']
        # IoU
        if iouThr is not None:
            t = np.where(iouThr == iouThrList)[0]
            s = s[t]
        s = s[:,:,:,aind, mind]
    else:
        # dimension of recall: [TxKxAxM]
        s = eval_dict['recall']
        if iouThr is not None:
            t = np.where(iouThr == iouThrList)[0]
            s = s[t]
        s = s[:,:,aind,mind]

    if len(s[s>-1])==0:
        mean_s = -1
    else:
        mean_s = np.mean(s[s>-1])

    per_class = []
    for i in range(s.shape[2]):
        si = s[:,:,i]
        per_class += [np.mean(si[si>-1])]

    if ap == 1:
        result_dict = {"AP%s" % (iouThr*100):np.array(per_class),
            "mAP%s"%(iouThr*100):mean_s}
    else:
        result_dict = {"RC%s" % (iouThr*100):np.array(per_class),
            "mRC%s"%(iouThr*100):mean_s}

    return result_dict

def evaluateImg(gt_dict,
                dt_dict,
                ious,
                imgId,
                catId,
                aRng,
                maxDet,
                iouThrs):
    '''
    perform evaluation for single category and image
    :return: dict (single image results)
    '''
    # p = self.params
    key = (imgId, catId)
    if key in gt_dict:
        gt = gt_dict[key]
    else:
        gt = []

    if key in dt_dict:
        dt = dt_dict[key]
    else:
        dt = []

    if len(gt) == 0 and len(dt) == 0:
        return None

    for g in gt:
        if (g['area'] < aRng[0] or g['area'] > aRng[1]):
            g['_ignore'] = 1
        else:
            g['_ignore'] = 0

    # sort dt highest score first, sort gt ignore last
    gtind = np.argsort([g['_ignore'] for g in gt], kind='mergesort')
    gt = [gt[i] for i in gtind]
    dtind = np.argsort([-d['score'] for d in dt], kind='mergesort')
    dt = [dt[i] for i in dtind[0:maxDet]]
    iscrowd = [int(o['iscrowd']) for o in gt]
    # load computed ious
    ious = (ious[(imgId, catId)][:, gtind]
            if len(ious[(imgId, catId)]) > 0 else ious[(imgId, catId)])

    T = len(iouThrs)
    G = len(gt)
    D = len(dt)
    gtm = np.zeros((T, G))
    dtm = np.zeros((T, D))
    gtIg = np.array([g['_ignore'] for g in gt])
    dtIg = np.zeros((T, D))
    if not len(ious) == 0:
        for tind, t in enumerate(iouThrs):
            for dind, d in enumerate(dt):
                # information about best match so far (m=-1 -> unmatched)
                iou = min([t, 1 - 1e-10])
                m = -1
                for gind, g in enumerate(gt):
                    # if this gt already matched, and not a crowd, continue
                    if gtm[tind, gind] > 0 and not iscrowd[gind]:
                        continue
                    # if dt matched to reg gt, and on ignore gt, stop
                    if m > -1 and gtIg[m] == 0 and gtIg[gind] == 1:
                        break
                    # continue to next gt unless better match made
                    if ious[(dind, gind)] < iou:
                        continue
                    # if match successful and best so far, store appropriately
                    iou = ious[(dind, gind)]
                    m = gind
                # if match made store id of match for both dt and gt
                if m == -1:
                    continue
                dtIg[tind, dind] = gtIg[m]
                dtm[tind, dind] = gt[m]['id']
                gtm[tind, m] = d['id']
    # set unmatched detections outside of area range to ignore
    a = np.array([d['area'] < aRng[0] or d['area'] > aRng[1]
                  for d in dt]).reshape((1, len(dt)))
    dtIg = np.logical_or(dtIg, np.logical_and(dtm == 0, np.repeat(a, T, 0)))
    # store results for given image and category
    return {
        'image_id': imgId,
        'category_id': catId,
        'aRng': aRng,
        'maxDet': maxDet,
        'dtIds': [d['id'] for d in dt],
        'gtIds': [g['id'] for g in gt],
        'dtMatches': dtm,
        'gtMatches': gtm,
        'dtScores': [d['score'] for d in dt],
        'gtIgnore': gtIg,
        'dtIgnore': dtIg,
    }


def accumulate(evalImgs, imgIds, catIds, iouThrs, recThrs,
               areaRng,
               maxDets):
    '''
    Accumulate per image evaluation results and store the result in self.eval
    :param p: input params for evaluation
    :return: None
    '''
    # print('Accumulating evaluation results...')

    # allows input customized parameters

    T = len(iouThrs)
    R = len(recThrs)
    K = len(catIds)
    A = len(areaRng)
    M = len(maxDets)

    precision = -np.ones(
        (T, R, K, A, M))  # -1 for the precision of absent categories
    recall = -np.ones((T, K, A, M))
    scores = -np.ones((T, R, K, A, M))

    # create dictionary for future indexing
    catIds = catIds
    setK = set(catIds)
    setA = set(map(tuple, areaRng))
    setM = set(maxDets)
    setI = set(imgIds)
    # get inds to evaluate
    k_list = [n for n, k in enumerate(catIds) if k in setK]
    m_list = [m for n, m in enumerate(maxDets) if m in setM]
    a_list = [
        n for n, a in enumerate(map(lambda x: tuple(x), areaRng)) if a in setA
    ]
    i_list = [n for n, i in enumerate(imgIds) if i in setI]
    I0 = len(imgIds)
    A0 = len(areaRng)
    # retrieve E at each category, area range, and max number of detections
    for k, k0 in enumerate(k_list):
        Nk = k0 * A0 * I0
        for a, a0 in enumerate(a_list):
            Na = a0 * I0
            for m, maxDet in enumerate(m_list):
                E = [evalImgs[Nk + Na + i] for i in i_list]
                E = [e for e in E if not e is None]
                if len(E) == 0:
                    continue
                dtScores = np.concatenate([e['dtScores'][0:maxDet] for e in E])

                # different sorting method generates slightly different results.
                # mergesort is used to be consistent as Matlab implementation.
                inds = np.argsort(-dtScores, kind='mergesort')
                dtScoresSorted = dtScores[inds]

                dtm = np.concatenate([e['dtMatches'][:, 0:maxDet] for e in E],
                                     axis=1)[:, inds]
                dtIg = np.concatenate([e['dtIgnore'][:, 0:maxDet] for e in E],
                                      axis=1)[:, inds]
                gtIg = np.concatenate([e['gtIgnore'] for e in E])
                npig = np.count_nonzero(gtIg == 0)
                if npig == 0:
                    continue
                tps = np.logical_and(dtm, np.logical_not(dtIg))
                fps = np.logical_and(np.logical_not(dtm), np.logical_not(dtIg))

                tp_sum = np.cumsum(tps, axis=1).astype(dtype=np.float)
                fp_sum = np.cumsum(fps, axis=1).astype(dtype=np.float)
                for t, (tp, fp) in enumerate(zip(tp_sum, fp_sum)):
                    tp = np.array(tp)
                    fp = np.array(fp)
                    nd = len(tp)
                    rc = tp / npig
                    pr = tp / (fp + tp + np.spacing(1))
                    q = np.zeros((R, ))
                    ss = np.zeros((R, ))

                    if nd:
                        recall[t, k, a, m] = rc[-1]
                    else:
                        recall[t, k, a, m] = 0

                    # numpy is slow without cython optimization for accessing elements
                    # use python array gets significant speed improvement
                    pr = pr.tolist()
                    q = q.tolist()

                    for i in range(nd - 1, 0, -1):
                        if pr[i] > pr[i - 1]:
                            pr[i - 1] = pr[i]

                    inds = np.searchsorted(rc, recThrs, side='left')
                    try:
                        for ri, pi in enumerate(inds):
                            q[ri] = pr[pi]
                            ss[ri] = dtScoresSorted[pi]
                    except:
                        pass
                    precision[t, :, k, a, m] = np.array(q)
                    scores[t, :, k, a, m] = np.array(ss)
    eval_dict = {
        'counts': [T, R, K, A, M],
        # 'date': datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
        'precision': precision,
        'recall': recall,
        'scores': scores,
    }

    return eval_dict


def targets_to_ann_list(images, targets):
    ann_list = []
    ann_id = 0

    img, targets = images[0], targets[0]
    image_id = targets["image_id"].item()
    img_dict = {}
    img_dict['id'] = image_id
    img_dict['height'] = img.shape[-2]
    img_dict['width'] = img.shape[-1]
    # dataset['images'].append(img_dict)
    bboxes = targets["boxes"]
    bboxes[:, 2:] -= bboxes[:, :2]
    bboxes = bboxes.tolist()
    labels = targets['labels'].tolist()
    areas = targets['area'].tolist()
    iscrowd = targets['iscrowd'].tolist()
    if 'masks' in targets:
        masks = targets['masks']
        # make masks Fortran contiguous for coco_mask
        masks = masks.permute(0, 2, 1).contiguous().permute(0, 2, 1)
    if 'keypoints' in targets:
        keypoints = targets['keypoints']
        keypoints = keypoints.reshape(keypoints.shape[0], -1).tolist()
    num_objs = len(bboxes)
    for i in range(num_objs):
        ann = {}
        ann['image_id'] = image_id
        ann['bbox'] = bboxes[i]
        ann['category_id'] = labels[i]
        # categories.add(labels[i])
        ann['score'] = 1
        ann['area'] = areas[i]
        ann['iscrowd'] = iscrowd[i]
        ann['id'] = ann_id
        if 'masks' in targets:
            ann["segmentation"] = mask_util.encode(masks[i].numpy())
        if 'keypoints' in targets:
            ann['keypoints'] = keypoints[i]
            ann['num_keypoints'] = sum(k != 0 for k in keypoints[i][2::3])
    
        ann_list += [ann]
        ann_id += 1

    return ann_list 
    
def preds_to_ann_list(preds_dict, mask_void=None):
    ann_list = []
    for original_id, prediction in preds_dict.items():
        if len(prediction) == 0:
            continue

        scores = prediction["scores"]
        labels = prediction["labels"]
        masks = prediction["masks"]

        masks = masks > 0.5

        scores = prediction["scores"].tolist()
        labels = prediction["labels"].tolist()

        rles = [] 
        for mask in masks:
            binmask = np.array(mask[0, :, :, np.newaxis], dtype=np.uint8, order="F")
            if mask_void is not None:
                ind = mask_void[0][:,:,None].numpy() == 1
                binmask[ind] = 0
            rle =  mask_util.encode(binmask)[0]
            rles += [rle]
        
        for rle in rles:
            rle["counts"] = rle["counts"].decode("utf-8")

        boxes = prediction["boxes"]
        boxes = xyxy_to_xywh(boxes).tolist()

        ann_list.extend(
            [
                {
                    "image_id": original_id,
                    "category_id": labels[k],
                    "bbox": boxes[k],
                    "segmentation": rle,
                    "score": scores[k],
                }
                for k, rle in enumerate(rles)
            ]
        )
    return ann_list


def targets_to_ann_list(images, targets):
    ann_list = []
    ann_id = 0

    img, targets = images[0], targets[0]
    image_id = targets["image_id"].item()
    img_dict = {}
    img_dict['id'] = image_id
    img_dict['height'] = img.shape[-2]
    img_dict['width'] = img.shape[-1]
    # dataset['images'].append(img_dict)
    bboxes = targets["boxes"]
    bboxes[:, 2:] -= bboxes[:, :2]
    bboxes = bboxes.tolist()
    labels = targets['labels'].tolist()
    areas = targets['area'].tolist()
    iscrowd = targets['iscrowd'].tolist()
    if 'masks' in targets:
        masks = targets['masks']
        # make masks Fortran contiguous for coco_mask
        masks = masks.permute(0, 2, 1).contiguous().permute(0, 2, 1)
    if 'keypoints' in targets:
        keypoints = targets['keypoints']
        keypoints = keypoints.reshape(keypoints.shape[0], -1).tolist()
    num_objs = len(bboxes)
    for i in range(num_objs):
        ann = {}
        ann['image_id'] = image_id
        ann['bbox'] = bboxes[i]
        ann['category_id'] = labels[i]
        # categories.add(labels[i])
        ann['score'] = 1
        ann['area'] = areas[i]
        ann['iscrowd'] = iscrowd[i]
        ann['id'] = ann_id
        if 'masks' in targets:
            ann["segmentation"] = mask_util.encode(masks[i].numpy())
        if 'keypoints' in targets:
            ann['keypoints'] = keypoints[i]
            ann['num_keypoints'] = sum(k != 0 for k in keypoints[i][2::3])
    
        ann_list += [ann]
        ann_id += 1

    return ann_list 

def bbox_yxyx_to_ann_list(bbox_yxyx, image_id=None, category_list=None, score_list=None):
    bbox_xywh = yxyx_to_xywh(bbox_yxyx)

    return bbox_to_ann_list(bbox_xywh, image_id=image_id, category_list=category_list, score_list=score_list)

def bbox_xyxy_to_ann_list(bbox_xyxy, image_id=None, category_list=None, score_list=None):
    bbox_xywh = xyxy_to_xywh(bbox_xyxy)

    return bbox_to_ann_list(bbox_xywh, image_id=image_id, category_list=category_list, score_list=score_list)

def bbox_to_ann_list(bbox_xywh, image_id=None, category_list=None, score_list=None):
    ann_list = []
    _, d = bbox_xywh.shape
    for i in range(bbox_xywh.shape[0]):
        bbox = bbox_xywh[i]
        x, y, w, h = bbox
        
        ann_list += [{
              "iscrowd":0,
              'bbox':[int(x), int(y), int(w), int(h)],
              "area":int(h*w),
             "image_id":image_id or 0,
             "category_id":category_list or 1,
             "height":int(h),
             "width":int(w),
             "score":score_list or 1}]

    return ann_list

def yxyx_to_xywh(boxes):
    ymin, xmin, ymax, xmax = boxes.unbind(1)
    return torch.stack((xmin, ymin, xmax - xmin, ymax - ymin), dim=1)

def xyxy_to_xywh(boxes):
    xmin, ymin, xmax, ymax = boxes.unbind(1)
    return torch.stack((xmin, ymin, xmax - xmin, ymax - ymin), dim=1)



from setuptools import setup
import setuptools

setup(name='haven-ai',
      version='0.3.1',
      description='Manage large-scale experiments',
      url='https://github.com/haven-ai/haven-ai',
      maintainer='Issam Laradji',
      maintainer_email='issam.laradji@gmail.com',
      license='MIT',
      scripts=['./haven_cli/haven'],
      packages=setuptools.find_packages(),
      zip_safe=False,
      python_requires='>=3.5',
      install_requires=[
        'requests>=0.0',
        'ipywidgets>=0.0',
        'tqdm>=0.0',
        'matplotlib>=0.0',
        'numpy>=0.0',
        'opencv-python-headless>=0.0',
        'pandas>=0.0',
        'Pillow>=0.0',
        'scikit-image>=0.0',
        'scikit-learn>=0.0',
        'scipy>=0.0',
        'sklearn>=0.0',
        'torch>=0.0',
        'torchvision>=0.0',
        'notebook >= 4.0',
        'jupyter>=0.0',
      ]),




from .. import haven_utils
from .. import haven_results as hr
from .. import haven_utils as hu 
from .. import haven_share as hd

import os
import pprint, json
import copy
import pprint
import pandas as pd 

try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except:
    print('widgets not available...')


def images_tab(self, output):
    tfigsize = widgets.Text(
        value=str(self.vars.get('figsize', '(10,5)')),
        description='figsize:',
        disabled=False
            )
    llegend_list = widgets.Text(
        value=str(self.vars.get('legend_list', '[model]')),
        description='legend_list:',
        disabled=False
            )
    
    t_n_images = widgets.Text(
        value=str(self.vars.get('n_images', '5')),
        description='n_images:',
        disabled=False
            )

    t_n_exps = widgets.Text(
        value=str(self.vars.get('n_exps', '3')),
        description='n_exps:',
        disabled=False
            )
    t_dirname = widgets.Text(
        value=str(self.vars.get('dirname', 'images')),
        description='dirname:',
        disabled=False
            )
    bdownload = widgets.Button(description="Download Images", 
                                layout=self.layout_button)
    bdownload_out = widgets.Output(layout=self.layout_button)
    brefresh = widgets.Button(description="Display Images")
    button = widgets.VBox([
            widgets.HBox([t_n_exps, t_n_images]),
            widgets.HBox([tfigsize, llegend_list, ]),
            widgets.HBox([t_dirname, ]),
            widgets.HBox([brefresh, bdownload, bdownload_out]),
                        ])

    output_plot = widgets.Output()

    with output:
        display(button)
        display(output_plot)

    def on_clicked(b):
        output_plot.clear_output()
        with output_plot:
            self.update_rm()
        
        
            w, h = tfigsize.value.strip('(').strip(')').split(',')
            self.vars['figsize'] = (int(w), int(h))
            self.vars['legend_list'] = hu.get_list_from_str(llegend_list.value)
            self.vars['n_images'] = int(t_n_images.value)
            self.vars['n_exps'] = int(t_n_exps.value)
            self.vars['dirname'] = t_dirname.value
            self.rm_original.fig_image_list =  self.rm.get_images(legend_list=self.vars['legend_list'], 
                    n_images=self.vars['n_images'],
                    n_exps=self.vars['n_exps'],
                    figsize=self.vars['figsize'],
                    dirname=self.vars['dirname'])
            show_inline_matplotlib_plots()
            
    brefresh.on_click(on_clicked)

    
    

    def on_download_clicked(b):
        fname = 'images'
        from matplotlib.backends.backend_pdf import PdfPages
        import matplotlib.pyplot as plt

        pp = PdfPages(fname)
        for fig in self.rm_original.fig_image_list:
            fig.savefig(pp, format='pdf')
        pp.close()

        bdownload_out.clear_output()
        
        with bdownload_out:
            display(FileLink(fname, result_html_prefix="Download: "))

    bdownload.on_click(on_download_clicked)


import copy
import glob
import os
import sys
import pprint
from itertools import groupby
from textwrap import wrap
import numpy as np
import pandas as pd
import pylab as plt
import tqdm


from .. import haven_jobs as hjb
from .. import haven_utils as hu
from .. import haven_share as hd


def get_plot(exp_list, savedir_base, 
             x_metric, y_metric,
             mode='line',
             filterby_list=None,
             title_list=None,
             legend_list=None,
             log_metric_list=None,
             figsize=None,
             avg_across=None,
             fig=None,
             axis=None,
             ylim=None,
             xlim=None,
             legend_fontsize=None,
             y_fontsize=None,
             x_fontsize=None,
             ytick_fontsize=None,
             xtick_fontsize=None,
             title_fontsize=None,
             legend_kwargs=None,
             map_title_list=tuple(),
             map_xlabel_list=tuple(),
             map_ylabel_list=dict(),
             bar_agg='min',
             verbose=True,
             show_legend=True,
             legend_format=None,
             title_format=None,
             cmap=None,
             show_ylabel=True,
             plot_confidence=True,
             x_cumsum=False,
             score_list_name='score_list.pkl',
             result_step=0,
             map_legend_list=dict()):
    """Plots the experiment list in a single figure.
    
    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    savedir_base : str
        A directory where experiments are saved
    x_metric : str
        Specifies metric for the x-axis
    y_metric : str
        Specifies metric for the y-axis
    title_list : [type], optional
        [description], by default None
    legend_list : [type], optional
        [description], by default None
    meta_list : [type], optional
        [description], by default None
    log_metric_list : [type], optional
        [description], by default None
    figsize : tuple, optional
        [description], by default (8, 8)
    avg_metric : [type], optional
        [description], by default None
    axis : [type], optional
        [description], by default None
    ylim : [type], optional
        [description], by default None
    xlim : [type], optional
        [description], by default None
    legend_fontsize : [type], optional
        [description], by default None
    y_fontsize : [type], optional
        [description], by default None
    ytick_fontsize : [type], optional
        [description], by default None
    xtick_fontsize : [type], optional
        [description], by default None
    legend_kwargs : [type], optional
        [description], by default None
    title_format: [str], optional
        [description], formatting of the title, by default None
    cmap: [str], optional
        [description], specify colormap, by default None
    
    Returns
    -------
    fig : [type]
        [description]
    axis : [type]
        [description]

    Examples
    --------
    >>> from haven import haven_results as hr
    >>> savedir_base='../results/isps/'
    >>> exp_list = hr.get_exp_list(savedir_base=savedir_base, 
    >>>                            filterby_list=[{'sampler':{'train':'basic'}}])
    >>> hr.get_plot(exp_list, savedir_base=savedir_base, x_metric='epoch', y_metric='train_loss', legend_list=['model'])
    """
    exp_list, style_list = hu.filter_exp_list(exp_list, filterby_list=filterby_list,
                             savedir_base=savedir_base, 
                             verbose=verbose,
                             return_style_list=True)
    # if len(exp_list) == 0:
    if axis is None:
        fig, axis = plt.subplots(nrows=1, ncols=1,
                                    figsize=figsize)
    # default properties
    if title_list is not None:
        title = get_label(title_list, exp_list[0], format_str=title_format)
    else:
        title = ''

    
    ylabel = y_metric
    xlabel = x_metric

    # map properties
    for map_dict in map_title_list:
        if title in map_dict:
            title = map_dict[title]

    for map_dict in map_xlabel_list:
        if x_metric in map_dict:
            xlabel = map_dict[x_metric]

    for map_dict in map_ylabel_list:
        if y_metric in map_dict:
            ylabel = map_dict[y_metric]

    # set properties
    axis.set_title(title, fontsize=title_fontsize)
    if ylim is not None:
        axis.set_ylim(ylim)
    if xlim is not None:
        axis.set_xlim(xlim)
    
    if log_metric_list and y_metric in log_metric_list:
        axis.set_yscale('log')
        ylabel = ylabel + ' (log)'
    
    if log_metric_list and x_metric in log_metric_list:
        axis.set_xscale('log')
        xlabel = xlabel + ' (log)'

    if show_ylabel:
        axis.set_ylabel(ylabel, fontsize=y_fontsize)

    axis.set_xlabel(xlabel, fontsize=x_fontsize)

    axis.tick_params(axis='x', labelsize=xtick_fontsize)
    axis.tick_params(axis='y', labelsize=ytick_fontsize)

    axis.grid(True)

    # if len(exp_list) > 50:
    #     if verbose:
    #         raise ValueError('many experiments in one plot...use filterby_list to reduce them')
    # if cmap is not None or cmap is not '':
    #     plt.rcParams["axes.prop_cycle"] = get_cycle(cmap)
    if mode == 'pretty_plot':
        tools.pretty_plot
    
    bar_count = 0
    visited_exp_ids = set()
    for exp_dict, style_dict in zip(exp_list, style_list):
        exp_id = hu.hash_dict(exp_dict)
        if exp_id in visited_exp_ids:
            continue

        savedir = os.path.join(savedir_base, exp_id)
        score_list_fname = os.path.join(savedir, score_list_name)

        if not os.path.exists(score_list_fname):
            if verbose:
                print('%s: %s does not exist...' % (exp_id,score_list_name))
            continue

        else:
            result_dict = get_result_dict(exp_dict, 
                            savedir_base, 
                            x_metric, 
                            y_metric,
                            plot_confidence=plot_confidence,
                            exp_list=exp_list, 
                            avg_across=avg_across,
                            verbose=verbose,
                            x_cumsum=x_cumsum,
                            score_list_name=score_list_name,
                            result_step=result_step)
            
            y_list = result_dict['y_list']
            x_list = result_dict['x_list']
            for eid in list(result_dict['visited_exp_ids']):
                visited_exp_ids.add(eid)
            if len(x_list) == 0 or np.array(y_list).dtype == 'object':
                x_list = np.NaN
                y_list = np.NaN
                if verbose:
                    print('%s: "(%s, %s)" not in score_list' % (exp_id, y_metric, x_metric))

            # map properties of exp
            if legend_list is not None:
                label = get_label(legend_list, exp_dict, format_str=legend_format)
            else:
                label = exp_id

            color = None
            marker = '*'
            linewidth = None
            markevery = None
            markersize = None

            if len(style_dict):
                marker = style_dict.get('marker', marker)
                label = style_dict.get('label', label)
                color = style_dict.get('color', color)
                linewidth = style_dict.get('linewidth', linewidth)
                markevery = style_dict.get('markevery', markevery)
                markersize = style_dict.get('markersize', markersize)

            if label in map_legend_list:
                label = map_legend_list[label]
            # plot
            if mode == 'pretty_plot':
                # plot the mean in a line
                # pplot = pp.add_yxList
                axis.plot(x_list, y_list, color=color, linewidth=linewidth, markersize=markersize,
                    label=str(label), marker=marker, markevery=markevery)
                # tools.pretty_plot
            elif mode == 'line':
                # plot the mean in a line
                axis.plot(x_list, y_list, color=color, linewidth=linewidth, markersize=markersize,
                    label=label, marker=marker, markevery=markevery)

                if avg_across and hasattr(y_list, 'size'):
                    # add confidence interval
                    axis.fill_between(x_list, 
                            y_list - result_dict.get('y_std_list', 0),
                            y_list + result_dict.get('y_std_list', 0), 
                            color = color,  
                            alpha=0.1)

            elif mode == 'bar':
                # plot the mean in a line
                if bar_agg == 'max':
                    y_agg = np.max(y_list)    
                elif bar_agg == 'min':
                    y_agg = np.min(y_list)
                elif bar_agg == 'mean':
                    y_agg = np.mean(y_list)
                elif bar_agg == 'last':
                    y_agg = [y for y in y_list if isinstance(y, float)][-1]
                    
                    
                width = 0.
                import math
                
                if math.isnan(y_agg):
                    s = 'NaN'
                    continue
                    
                else:
                    s="%.3f" % y_agg

                axis.bar([bar_count + width], 
                        [y_agg],
                        color=color,
                        label=label,
                        # label='%s - (%s: %d, %s: %.3f)' % (label, x_metric, x_list[-1], y_metric, y_agg)
                        )
                if color is not None:
                    bar_color = color
                else:
                    bar_color = 'black'

                # minimum, maximum = axis.get_ylim()
                # y_height = .05 * (maximum - minimum)

                # axis.text(bar_count, y_agg + .01, "%.3f"%y_agg, color=bar_color, fontweight='bold')
                axis.text(x=bar_count, y = y_agg*1.01, 
                            s=s, 
                            fontdict=dict(fontsize=(y_fontsize or 12)), color='black', 
                            fontweight='bold')
                axis.set_xticks([])
                bar_count += 1
            else:
                raise ValueError('mode %s does not exist. Options: (line, bar)' % mode)

    legend_kwargs = legend_kwargs or {"loc":2, "bbox_to_anchor":(1.05,1),
                                      'borderaxespad':0., "ncol":1}
    
    if mode == 'pretty_plot':
        pass
    elif show_legend:
        axis.legend(fontsize=legend_fontsize, **legend_kwargs)

    plt.tight_layout()
    

    return fig, axis

def get_label(original_list, exp_dict, format_str=None):
    label_list = []
    for k in original_list:
        depth_list = k.split('.')
        sub_dict = exp_dict
        for d in depth_list:
            if sub_dict is None or d not in sub_dict:
                sub_dict = None
                break
            sub_dict = sub_dict[d]
            
        label_list += [str(sub_dict)]
        
    if format_str:
        label = format_str.format(*label_list)
    else:
        label = '_'.join(label_list)
    
    label = '\n'.join(wrap(label, 40))
    return label


def get_result_dict(exp_dict, 
                    savedir_base, 
                    x_metric, 
                    y_metric,
                    exp_list=None, 
                    avg_across=False,
                    verbose=False,
                    plot_confidence=True,
                    x_cumsum=False,
                    score_list_name='score_list.pkl',
                    result_step=0):
    visited_exp_ids = set()
    exp_id = hu.hash_dict(exp_dict)
    savedir = os.path.join(savedir_base, exp_id)
    score_list_fname = os.path.join(savedir, score_list_name)

    # get scores
    if not avg_across:
        # get score list
        score_list = hu.load_pkl(score_list_fname)
        x_list = []
        y_list = []
        for score_dict in score_list:
            if x_metric in score_dict and y_metric in score_dict:
                x_list += [score_dict[x_metric]]
                y_list += [score_dict[y_metric]]
        y_std_list = []

    else:
        assert exp_list is not None, 'exp_list must be passed'
        # average score list across an hparam
        
        filter_dict = {k:exp_dict[k] for k in exp_dict if k not in avg_across}
        exp_sublist = hu.filter_exp_list(exp_list, 
                                        filterby_list=[filter_dict], 
                                        savedir_base=savedir_base,
                                        verbose=verbose)
        def count(d):
            return sum([count(v) if isinstance(v, dict) 
                            else 1 for v in d.values()])
        n_values = count(filter_dict) + 1
        exp_sublist = [sub_dict for sub_dict in exp_sublist
                            if n_values == count(sub_dict)]
        # get score list
        x_dict = {}
        
        uniques= np.unique([sub_dict[avg_across] for sub_dict in exp_sublist])
        # print(uniques, len(exp_sublist))
        assert(len(exp_sublist)>0)
        assert(len(uniques) == len(exp_sublist))
        for sub_dict in exp_sublist:
            sub_id = hu.hash_dict(sub_dict)
            sub_score_list_fname = os.path.join(savedir_base, sub_id, score_list_name)

            if not os.path.exists(sub_score_list_fname):
                if verbose:
                    print('%s: %s does not exist...' % (sub_id, score_list_name))
                continue

            visited_exp_ids.add(sub_id)

            sub_score_list = hu.load_pkl(sub_score_list_fname)

            for score_dict in sub_score_list:
                if x_metric in score_dict and y_metric in score_dict:
                    x_val = score_dict[x_metric]
                    if not x_val in x_dict:
                        x_dict[x_val] = []

                    x_dict[x_val] += [score_dict[y_metric]]
        # import ipdb; ipdb.set_trace()
        if len(x_dict) == 0:
            x_list = []
            y_list = []
        else:
            x_list = np.array(list(x_dict.keys()))
            y_list_raw = list(x_dict.values())
            y_list_raw = [yy for yy in y_list_raw if len(yy) == len(exp_sublist)]
            y_list_all = np.array(y_list_raw)
            x_list = x_list[:len(y_list_all)]
            if y_list_all.dtype == 'object' or len(y_list_all)==0:
                x_list = []
                y_list = []
                y_std_list = []
            else:
                if plot_confidence:
                    y_std_list = np.std(y_list_all, axis=1)
                else:
                    y_std_list = 0
                y_list = np.mean(y_list_all, axis=1)  
                
    if x_cumsum:
        x_list = np.cumsum(x_list)
    
    if result_step == 0:
        return {'y_list':y_list, 
                'x_list':x_list,
                'y_std_list':y_std_list,
                'visited_exp_ids':visited_exp_ids}
    else:
        return {'y_list':y_list[::result_step], 
                'x_list':x_list[::result_step],
                'y_std_list':y_std_list,
                'visited_exp_ids':visited_exp_ids}


import copy
import glob
import os
import sys
import pprint
from itertools import groupby
from textwrap import wrap
import numpy as np
import pandas as pd
import pylab as plt
import tqdm

from .. import haven_jobs as hjb
from .. import haven_utils as hu
from .. import haven_share as hd
from .tables_scores import *
from .plots_line import *
from . import images_fig
# from . import tools


class ResultManager:
    def __init__(self, 
                 savedir_base,
                 exp_list=None,
                 filterby_list=None,
                 verbose=True,
                 has_score_list=False,
                 exp_groups=None,
                 mode_key=None,
                 exp_ids=None,
                 save_history=False,
                 score_list_name='score_list.pkl',
                 account_id=None):
        """[summary]
        
        Parameters
        ----------
        savedir_base : [type]
            A directory where experiments are saved
        exp_list : [type], optional
            [description], by default None
        filterby_list : [type], optional
            [description], by default None
        has_score_list : [type], optional
            [description], by default False
        
        Example
        -------
        >>> from haven import haven_results as hr
        >>> savedir_base='../results'
        >>> rm = hr.ResultManager(savedir_base=savedir_base,
                                filterby_list=[{'dataset':'mnist'}],
                                verbose=1)
        >>> for df in rm.get_score_df():
        >>>     display(df)
        >>> fig_list = rm.get_plot_all(y_metric_list=['train_loss', 'val_acc'], 
                                    order='groups_by_metrics',
                                    x_metric='epoch', 
                                    figsize=(15,6),
                                    title_list=['dataset'],
                                    legend_list=['model']) 
        """
        
        # sanity checks
        assert os.path.exists(savedir_base), '%s does not exist' % savedir_base

        self.exp_groups = {}
        if exp_groups is not None:
            if isinstance(exp_groups, dict):
                # load from a dict
                self.exp_groups = exp_groups
            elif os.path.exists(exp_groups):
                # load from a file
                self.exp_groups = hu.load_py(exp_groups).EXP_GROUPS
            else:
                raise ValueError('%s does not exist...' % exp_groups)
        
        # rest
        self.score_list_name = score_list_name
        self.mode_key = mode_key
        self.has_score_list = has_score_list
        self.save_history = save_history
        self.account_id = account_id
        # get exp _list
        
        if exp_ids is not None:
            assert exp_list is None, "settings exp_ids require exp_list=None"
            assert exp_groups is None, "settings exp_ids require exp_groups=None"
            exp_list = []
            for exp_id in exp_ids:
                exp_list += [hu.load_json(os.path.join(savedir_base, exp_id, 'exp_dict.json'))]
        else:
            if exp_list is None:
                exp_list = hu.get_exp_list(savedir_base=savedir_base, verbose=verbose)
            else:
                exp_list = exp_list
            
            if len(exp_list) == 0:
                raise ValueError('exp_list is empty...')

        exp_list_with_scores = [e for e in exp_list if 
                                    os.path.exists(os.path.join(savedir_base, 
                                                                hu.hash_dict(e),
                                                                score_list_name))]
        if has_score_list:
            exp_list = exp_list_with_scores

        self.exp_list_all = copy.deepcopy(exp_list)

        
        self.score_keys  = ['None']

        if len(exp_list_with_scores):
            score_fname = os.path.join(savedir_base, hu.hash_dict(exp_list_with_scores[0]), score_list_name)
            self.score_keys = ['None'] + list(hu.load_pkl(score_fname)[0].keys())
                    
                                                        
        self.savedir_base = savedir_base
        
        self.filterby_list = filterby_list
        self.verbose = verbose

        self.n_exp_all = len(exp_list)

        self.exp_list = hu.filter_exp_list(exp_list, 
                                        filterby_list=filterby_list, 
                                        savedir_base=savedir_base,
                                        verbose=verbose)

        if len(self.exp_list) != 0:
            self.exp_params = list(self.exp_list[0].keys())
        else:
            self.exp_params = []

        
        
        if mode_key:
            for exp_dict in exp_list:
                exp_dict[mode_key] = 1
            for exp_dict in self.exp_list_all:
                exp_dict[mode_key] = 1

        self.exp_groups['all'] = copy.deepcopy(self.exp_list_all)
    
    def get_state_dict(self):
        pass
    def load_state_dict(self, state_dict):
        pass

    def get_plot(self, groupby_list=None, savedir_plots=None, filterby_list=None, **kwargs):
        fig_list = []
        filterby_list = filterby_list or self.filterby_list
        exp_groups = hu.group_exp_list(self.exp_list, groupby_list)
        
        for i, exp_list in enumerate(exp_groups):
            fig, ax = get_plot(exp_list=exp_list, savedir_base=self.savedir_base, filterby_list=filterby_list, 
                        
                        verbose=self.verbose,
                        score_list_name=self.score_list_name,
                               **kwargs)
            fig_list += [fig]

            # save image if given
            if savedir_plots != '' and savedir_plots is not None:
                os.makedirs(savedir_plots, exist_ok=True)
                save_fname = os.path.join(savedir_plots, "%d.png" % i )
                fig.savefig(save_fname, bbox_inches='tight')

        return fig_list

    def get_plot_all(self, y_metric_list, order='groups_by_metrics', 
                     groupby_list=None, ylim_list=None, xlim_list=None,
                     savedir_plots=None, legend_last_row_only=False, show_legend_all=None,
                     **kwargs):
        """[summary]
        
        Parameters
        ----------
        y_metric_list : [type]
            [description]
        order : str, optional
            [description], by default 'groups_by_metrics'
        
        Returns
        -------
        [type]
            [description]
        
        """
        
        if order not in ['groups_by_metrics', 'metrics_by_groups']:
            raise ValueError('%s order is not defined, choose between %s' % (order, ['groups_by_metrics', 'metrics_by_groups']))
        exp_groups = hu.group_exp_list(self.exp_list, groupby_list)
        figsize = kwargs.get('figsize') or None
        
        fig_list = []

        if not isinstance(y_metric_list, list):
            y_metric_list = [y_metric_list]

        if ylim_list is not None:
            assert isinstance(ylim_list[0], list), "ylim_list has to be lists of lists"
        if xlim_list is not None:
            assert isinstance(xlim_list[0], list), "xlim_list has to be lists of lists"

        if order == 'groups_by_metrics':
            for j, exp_list in enumerate(exp_groups):   
                fig, ax_list = plt.subplots(nrows=1, ncols=len(y_metric_list), figsize=figsize)
                if not hasattr(ax_list, 'size'):
                    ax_list = [ax_list]
                for i, y_metric in enumerate(y_metric_list):
                    if i == (len(y_metric_list) - 1):
                        show_legend = True
                    else:
                        show_legend = False
                    
                    ylim = None
                    xlim = None
                    if ylim_list is not None:
                        assert len(ylim_list) == len(exp_groups), "ylim_list has to have %d rows" % len(exp_groups)
                        assert len(ylim_list[0]) == len(y_metric_list), "ylim_list has to have %d cols" % len(y_metric_list)
                        ylim = ylim_list[j][i]
                    if xlim_list is not None:
                        assert len(xlim_list) == len(exp_groups), "xlim_list has to have %d rows" % len(exp_groups)
                        assert len(xlim_list[0]) == len(y_metric_list), "xlim_list has to have %d cols" % len(y_metric_list)
                        xlim = xlim_list[j][i]
                    if show_legend_all is not None:
                        show_legend = show_legend_all
                    fig, _ = get_plot(exp_list=exp_list, savedir_base=self.savedir_base, y_metric=y_metric, 
                                    fig=fig, axis=ax_list[i], verbose=self.verbose, filterby_list=self.filterby_list,
                                    show_legend=show_legend,
                                    ylim=ylim, xlim=xlim,
                                    score_list_name=self.score_list_name,
                                    **kwargs)
                fig_list += [fig]

        elif order == 'metrics_by_groups':

            for j, y_metric in enumerate(y_metric_list):   
                fig, ax_list = plt.subplots(nrows=1, ncols=len(exp_groups) , figsize=figsize)
                if not hasattr(ax_list, 'size'):
                    ax_list = [ax_list]
                for i, exp_list in enumerate(exp_groups):
                    if i == 0:
                        show_ylabel = True
                    else:
                        show_ylabel = False

                    if i == (len(exp_groups) - 1):
                        show_legend = True
                    else:
                        show_legend = False

                    if legend_last_row_only and j < (len(y_metric_list) - 1):
                        show_legend = False

                    ylim = None
                    xlim = None
                    if ylim_list is not None:
                        assert len(ylim_list) == len(y_metric_list), "ylim_list has to have %d rows" % len(exp_groups)
                        assert len(ylim_list[0]) == len(exp_groups), "ylim_list has to have %d cols" % len(y_metric_list)
                        ylim = ylim_list[j][i]
                    if xlim_list is not None:
                        assert len(xlim_list) == len(y_metric_list), "xlim_list has to have %d rows" % len(exp_groups)
                        assert len(xlim_list[0]) == len(exp_groups), "xlim_list has to have %d cols" % len(y_metric_list)
                        xlim = xlim_list[j][i]

                    if show_legend_all is not None:
                        show_legend = show_legend_all
                        
                    fig, _ = get_plot(exp_list=exp_list, savedir_base=self.savedir_base, y_metric=y_metric, 
                                    fig=fig, axis=ax_list[i], verbose=self.verbose, filterby_list=self.filterby_list,
                                    ylim=ylim, xlim=xlim,
                                    show_legend=show_legend,
                                    show_ylabel=show_ylabel,
                                    score_list_name=self.score_list_name,
                                    **kwargs)
                fig_list += [fig]

        plt.tight_layout()
        if savedir_plots:
            for i in range(len(fig_list)):
                os.makedirs(savedir_plots, exist_ok=True)
                fname = os.path.join(savedir_plots + '_%d.pdf' % i)
                fig_list[i].savefig(fname,
                                     dpi=300, 
                                     bbox_inches='tight')
                print(fname, 'saved')
                
        return fig_list
    
    def get_score_df(self, **kwargs):
        """[summary]
        
        Returns
        -------
        [type]
            [description]
        """
        df_list = get_score_df(exp_list=self.exp_list, 
                    savedir_base=self.savedir_base, verbose=self.verbose, 
                    score_list_name=self.score_list_name,
                    **kwargs)
        return df_list 

    def to_dropbox(self, outdir_base, access_token):
        """[summary]
        """ 
        hd.to_dropbox(self.exp_list, savedir_base=self.savedir_base, 
                          outdir_base=outdir_base,
                          access_token=access_token)

    def get_exp_list_df(self, **kwargs):
        """[summary]
        
        Returns
        -------
        [type]
            [description]
        """
        df_list = get_exp_list_df(exp_list=self.exp_list,
                     verbose=self.verbose, **kwargs)
        
        return df_list 

    def get_exp_table(self, **kwargs):
        """[summary]
        
        Returns
        -------
        [type]
            [description]
        """
        table = get_exp_list_df(exp_list=self.exp_list, verbose=self.verbose, **kwargs)
        return table 

     
    def get_score_table(self, **kwargs):
        """[summary]
        
        Returns
        -------
        [type]
            [description]
        """
        table = get_score_df(exp_list=self.exp_list, 
                    savedir_base=self.savedir_base, 
                    score_list_name=self.score_list_name,
                    filterby_list=self.filterby_list,
                    verbose=self.verbose, **kwargs)
        return table 

    def get_score_lists(self, **kwargs):
        """[summary]
        
        Returns
        -------
        [type]
            [description]
        """
        score_lists = get_score_lists(exp_list=self.exp_list, 
                                      savedir_base=self.savedir_base, 
                                      score_list_name=self.score_list_name,
                                      filterby_list=self.filterby_list,
                                      verbose=self.verbose, **kwargs)
        return score_lists

    def get_images(self, **kwargs):
        """[summary]
        
        Returns
        -------
        [type]
            [description]
        """
        return images_fig.get_images(exp_list=self.exp_list, savedir_base=self.savedir_base, verbose=self.verbose, **kwargs)

    def get_job_summary(self, columns=None, add_prefix=False, **kwargs):
        """[summary]
        """
        exp_list = hu.filter_exp_list(self.exp_list, self.filterby_list, savedir_base=self.savedir_base, verbose=self.verbose)
        jm = hjb.JobManager(exp_list=exp_list, savedir_base=self.savedir_base, account_id=self.account_id, **kwargs)
        summary_list = jm.get_summary_list(columns=columns, add_prefix=add_prefix)

        return summary_list
            
    def to_zip(self, savedir_base='', fname='tmp.zip', **kwargs):
        """[summary]
        
        Parameters
        ----------
        fname : [type]
            [description]
        """
        from haven.haven_share import haven_dropbox as hd
        if savedir_base == '':
            savedir_base = self.savedir_base
        exp_id_list = [hu.hash_dict(exp_dict) for exp_dict in self.exp_list]
        hd.zipdir(exp_id_list, savedir_base, fname, **kwargs)

    def to_dropbox(self, fname, dropbox_path=None, access_token=None):
        """[summary]
        
        Parameters
        ----------
        fname : [type]
            [description]
        dropbox_path : [type], optional
            [description], by default None
        access_token : [type], optional
            [description], by default None
        """
        from haven import haven_dropbox as hd

        out_fname = os.path.join(dropbox_path, fname)
        src_fname = os.path.join(self.savedir_base, fname)
        self.to_zip(src_fname)
        hd.upload_file_to_dropbox(src_fname, out_fname, access_token)
        print('saved: https://www.dropbox.com/home/%s' % out_fname)






